from dataclasses import dataclass
from typing import Tuple, List, Optional, Dict, Any, Type, Collection

from pathlib import Path

from data_to_paper.base_steps import DebuggerConverser
from data_to_paper.base_steps.request_code import CodeReviewPrompt
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.code_and_output_files.output_file_requirements import \
    OutputFileRequirements, PickleContentOutputFileRequirement
from data_to_paper.code_and_output_files.referencable_text import LabeledNumericReferenceableText, \
    convert_str_to_latex_label
from data_to_paper.llm_coding_utils import describe_df, df_to_figure, df_to_latex
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.check_df_to_funcs.df_checker import check_analysis_df
from data_to_paper.research_types.hypothesis_testing.env import get_max_rows_and_columns
from data_to_paper.run_gpt_code.code_runner import CodeRunner
from data_to_paper.run_gpt_code.overrides.dataframes.df_with_attrs import InfoDataFrameWithSaveObjFuncCall
from data_to_paper.research_types.hypothesis_testing.coding.analysis.utils import get_pickle_dump_attr_replacer
from data_to_paper.research_types.hypothesis_testing.coding.base_code_conversers import BaseTableCodeProductsGPT
from data_to_paper.research_types.hypothesis_testing.coding.utils import create_pandas_and_stats_contexts
from data_to_paper.research_types.hypothesis_testing.scientific_products import HypertargetPrefix
from data_to_paper.run_gpt_code.extract_and_check_code import ModifyAndCheckCodeExtractor, CodeExtractor
from data_to_paper.run_gpt_code.overrides.dataframes.utils import df_to_string_with_format_value, \
    format_numerics_and_iterables
from data_to_paper.run_gpt_code.overrides.pvalue import is_containing_p_value, OnStr, OnStrPValue
from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceDict


@dataclass(frozen=True)
class BaseDataFramePickleContentOutputFileRequirement(PickleContentOutputFileRequirement):
    referenceable_text_cls: type = LabeledNumericReferenceableText
    output_folder: Optional[Path] = None  # where the figures will be saved

    def _is_figure(self, content: InfoDataFrameWithSaveObjFuncCall) -> bool:
        return content.get_func_call().func == df_to_figure

    def _plot_kind(self, content: InfoDataFrameWithSaveObjFuncCall) -> Optional[str]:
        if not self._is_figure(content):
            return None
        return content.get_func_call().kwargs.get('kind')

    def _get_content_and_header_for_product(
            self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.PRODUCT):
        func_name = content.get_func_call().func_name
        content, header = super()._get_content_and_header_for_product(content, filename, num_file, level, view_purpose)
        header += f' (with {func_name})'
        return content, header

    def _convert_content_to_labeled_text(self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None,
                                         num_file: int = 0, view_purpose: ViewPurpose = None) -> str:
        with OnStrPValue(self._convert_view_purpose_to_pvalue_on_str(view_purpose)):
            return describe_df(
                content, should_format=True,
                max_rows_and_columns_to_show=get_max_rows_and_columns(is_figure=self._is_figure(content),
                                                                      kind=self._plot_kind(content),
                                                                      to_show=True))

    def _get_content_and_header_for_app_html(
            self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.APP_HTML):
        func_call = content.get_func_call()
        with OnStrPValue(self._convert_view_purpose_to_pvalue_on_str(view_purpose)):
            html = func_call.call(is_html=True, figure_folder=self.output_folder)
        return html, f'<h{level}>{filename}</h{level}>'

    def get_code_line_str_for_file(self, filename: str, content: InfoDataFrameWithSaveObjFuncCall) -> Optional[str]:
        func_call = content.get_func_call()
        func = func_call.func
        table_or_figure = 'Table' if func == df_to_latex else 'Figure'
        return f'## {table_or_figure} {func_call.filename}:'

    def get_hyperlink_label_for_file_header(self, filename: str,
                                            content: InfoDataFrameWithSaveObjFuncCall) -> Optional[str]:
        source_file = content.get_prior_filename()
        if source_file:
            return convert_str_to_latex_label(source_file + '.pkl', prefix='file')
        return super().get_hyperlink_label_for_file_header(filename, content)

    def get_issues_for_output_file_content(self, filename: str, content: Any) -> List[RunIssue]:
        issues = super().get_issues_for_output_file_content(filename, content)
        issues.extend(self._check_df(content))
        return issues

    def _check_df(self, content: InfoDataFrameWithSaveObjFuncCall) -> List[RunIssue]:
        return []


@dataclass(frozen=True)
class DataFramePickleContentOutputFileRequirement(BaseDataFramePickleContentOutputFileRequirement):
    def _check_df(self, content: InfoDataFrameWithSaveObjFuncCall) -> List[RunIssue]:
        return check_analysis_df(content, output_folder=self.output_folder)

    def _get_content_and_header_for_final_appendix(
            self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.FINAL_APPENDIX):
        with OnStrPValue(OnStr.WITH_ZERO):
            content = df_to_string_with_format_value(content)
        return content, f'% {filename}'


class DictPickleContentOutputFileRequirement(PickleContentOutputFileRequirement):

    def _convert_content_to_labeled_text(self, content: Any, filename: str = None, num_file: int = int,
                                         view_purpose: ViewPurpose = None) -> str:
        content = NiceDict(content, format_numerics_and_iterables=format_numerics_and_iterables)
        return super()._convert_content_to_labeled_text(content, filename, num_file, view_purpose)

    def get_code_line_str_for_file(self, filename: str, content: Any) -> Optional[str]:
        return '# SAVE ADDITIONAL RESULTS'

    def get_issues_for_output_file_content(self, filename: str, content: Any) -> List[RunIssue]:
        # all checks are done in _pickle_dump_with_checks
        return super().get_issues_for_output_file_content(filename, content)


class StaticChecks(ModifyAndCheckCodeExtractor):
    class_and_from_formula: Tuple[str, str] = (
        ('GLS', 'gls'),
        ('WLS', 'wls'),
        ('OLS', 'ols'),
        ('GLSAR', 'glsar'),
        ('MixedLM', 'mixedlm'),
        ('GLM', 'glm'),
        ('RLM', 'rlm'),
        ('MNLogit', 'mnlogit'),
        ('Logit', 'logit'),
        ('Probit', 'probit'),
        ('Poisson', 'poisson'),
        ('NegativeBinomial', 'negativebinomial'),
        ('QuantReg', 'quantreg'),
        ('PHReg', 'phreg'),
        ('OrdinalGEE', 'ordinal_gee'),
        ('NominalGEE', 'nominal_gee'),
        ('GEE', 'gee'),
        ('GLMGam', 'glmgam'),
        ('ConditionalLogit', 'conditional_logit'),
        ('ConditionalMNLogit', 'conditional_mnlogit'),
        ('ConditionalPoisson', 'conditional_poisson'),
    )

    def get_issues_for_static_code_check(self, code: str) -> List[RunIssue]:
        issues = super().get_issues_for_static_code_check(code)

        for class_name, from_formula in self.class_and_from_formula:
            if class_name + '(' in code:
                issues.append(RunIssue(
                    category='Coding: good practices',
                    issue=f'You are using the `{class_name}` class. ',
                    instructions=dedent_triple_quote_str(f"""
                        You should use the `{from_formula}` function instead, so that the formula is clearly \t
                        specified as a string. 
                        Reminder: For interactions, if any, use the `*` operator in the formula, rather than \t
                        manually multiplying the variables.
                        """),
                    code_problem=CodeProblem.StaticCheck,
                ))

        return issues


@dataclass
class DataAnalysisDebuggerConverser(DebuggerConverser):

    def _get_issues_for_created_output_files(self, code_and_output: CodeAndOutput, contexts) -> List[RunIssue]:
        """
        Check that a PValue instance appear in at least one of the created tables.
        """
        issues = super()._get_issues_for_created_output_files(code_and_output, contexts)
        if issues:
            return issues
        issues = self.get_issues_for_missing_p_values(code_and_output)
        if issues:
            return issues
        return self._get_issues_for_df_comments(code_and_output, contexts)

    def get_issues_for_missing_p_values(self, code_and_output: CodeAndOutput) -> List[RunIssue]:
        issues = []
        any_pvalues = False
        for names_to_contents in code_and_output.created_files.values():
            for content in names_to_contents.values():
                if is_containing_p_value(content):
                    any_pvalues = True
                    break
        if not any_pvalues:
            issues.append(RunIssue(
                category='Statistics: good practices',
                issue='We are presenting results for a statistical-testing paper, but no p-values are reported in '
                      'any of the created files.',
                instructions='Please revise the code to perform statistical tests and report p-values in the tables.',
                code_problem=CodeProblem.OutputFileContentA,
            ))
        return issues

    def _get_issues_for_df_comments(self, code_and_output: CodeAndOutput, contexts) -> List[RunIssue]:
        issues = []
        files_to_requirements_and_contents = \
            code_and_output.created_files.get_created_files_to_requirements_and_contents(
                match_filename='df_*.pkl', is_content=True)
        for df_file_name, (requirement, content) in files_to_requirements_and_contents.items():
            df_header = requirement.get_code_line_str_for_file(df_file_name, content)
            if df_header and df_header not in code_and_output.code:
                issues.append(RunIssue(
                    category="Code structure",
                    issue=f'Your code is missing a comment "{df_header}".',
                    instructions=dedent_triple_quote_str("""
                        Please make sure all saved tables/figures have a comment with their chosen tag.
                        This comment should be placed at the beginning of the section of the code that creates \t
                        the table/figure. Like this:

                        ```python
                        ...

                        ## Table df_tag1:
                        caption = "..."
                        ...  # any analysis code related to df_tag1
                        df_tag1 = ...
                        df_to_latex(df_tag1, 'df_tag1', caption=caption)

                        ## Figure df_tag2:
                        caption = "..."
                        ...  # any analysis code related to df_tag2
                        df_tag2 = ...
                        df_to_figure(df_tag2, 'df_tag2', caption=caption)

                        ...
                        ```

                        If you are creating multiple tables in the same code section, '
                        you should precede this section with a separate comment for each of the tables.
                        Like this:

                        ```python
                        ...

                        ## Table df_tag1:
                        caption1 = "..."                        
                        ## Figure df_tag2:
                        caption2 = "..."

                        ...  # any analysis code related to both df_tag1 and df_tag2
                        df_tag1 = ...
                        df_to_latex(df_tag1, 'df_tag1', caption=caption1)
                        df_tag2 = ...
                        df_to_figure(df_tag2, 'df_tag2', caption=caption2)

                        ...
                        ```
                        """),
                    code_problem=CodeProblem.OutputFileAnnotation,
                ))
        return issues


@dataclass
class AnalysisCodeRunner(CodeRunner):
    modified_imports: Tuple[Tuple[str, Optional[str]]] = CodeRunner.modified_imports + (
        ('my_utils', 'data_to_paper.research_types.hypothesis_testing.coding.analysis.my_utils'),
    )


@dataclass
class DataAnalysisCodeProductsGPT(BaseTableCodeProductsGPT):
    code_step: str = 'data_analysis'
    code_extractor_cls: Type[CodeExtractor] = StaticChecks
    debugger_cls: Type[DebuggerConverser] = DataAnalysisDebuggerConverser
    code_runner_cls: Type[CodeRunner] = AnalysisCodeRunner
    headers_required_in_code: Tuple[str, ...] = (
        '# IMPORT',
        '# LOAD DATA',
        '# DATASET PREPARATIONS',
        '# DESCRIPTIVE STATISTICS',
        '# PREPROCESSING',
        '# ANALYSIS',
        '# SAVE ADDITIONAL RESULTS',
    )
    max_debug_iterations_per_attempt: int = 20
    max_code_revisions: int = 3
    user_agent: ScientificAgent = ScientificAgent.Debugger

    background_product_fields: Tuple[str, ...] = \
        ('data_file_descriptions', 'outputs:data_exploration', 'codes:data_preprocessing',
         'created_files_headers:data_preprocessing', 'research_goal', 'hypothesis_testing_plan')
    background_product_fields_to_hide_during_code_revision: Tuple[str, ...] = \
        ('outputs:data_exploration', 'research_goal')
    allow_data_files_from_sections: Tuple[Optional[str]] = (None, 'data_preprocessing')

    supported_packages: Tuple[str, ...] = ('pandas', 'numpy', 'scipy', 'statsmodels', 'sklearn', 'pickle')

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements(
            [DataFramePickleContentOutputFileRequirement('df_*.pkl', 1, output_folder=self.output_directory),
             DictPickleContentOutputFileRequirement('additional_results.pkl', 1,
                                                    hypertarget_prefixes=HypertargetPrefix.ADDITIONAL_RESULTS.value)
             ])

    provided_code: str = dedent_triple_quote_str('''
            {df_to_latex_doc}

            {df_to_figure_doc}
        ''')

    df_to_latex_extra_vars: str = ''
    df_to_latex_extra_vars_explain: str = ''
    df_to_figure_extra_latex_vars: str = ''
    df_to_figure_extra_latex_vars_explain: str = ''
    df_to_figure_extra_plot_vars: str = ''
    df_to_figure_extra_plot_vars_explain: str = ''

    mission_prompt: str = dedent_triple_quote_str("""
        Write a complete Python code to analyze the data according to our \t
        "{research_goal}" and "{hypothesis_testing_plan}". 

        The code should create scientific Tables and Figures for our paper. 
        It should use the following provided functions:

        ```python
        {provided_code}
        ```

        The code must have the following structure (with these exact capitalized headers):

        `# IMPORT`
        `from my_utils import df_to_latex, df_to_figure`
        `import pickle`
        You can also import here any other packages including: 
        {supported_packages}\t

        `# LOAD DATA`
        Load the data from the original data files described above (see "{data_file_descriptions}").
        {list_additional_data_files_if_any}\t

        `# DATASET PREPARATIONS`
        * Join data files as needed.
        * Dealing with missing, unknown, or undefined values, or with special numeric values that stand for \t
        unknown/undefined (check in the "{data_file_descriptions}" for any such values, and \t
        consider also the "{outputs:data_exploration}").
        * Create new variables as needed.
        * Restrict the data based on exclusion/inclusion criteria (to match study goal, if applicable).
        * Standardize numeric values with different units into same-unit values.

        If no dataset preparations are needed, write below this header:
        `# No dataset preparations are needed.`

        `# DESCRIPTIVE STATISTICS`
        * In light of our study goals and the hypothesis testing plan (see above "{research_goal}" and \t
        "{hypothesis_testing_plan}"), decide whether and which descriptive statistics displayitems are needed.

        - If you decide that no descriptive statistics is needed, write:
        `# No descriptive statistics table is needed because <your reasons here>.` 

        - For descriptive statistics Table:
        `## Table df_desc_stat:`
        `caption = "Descriptive statistics of ..."`
        Write here the code to create a descriptive statistics dataframe `df_desc_stat`.
        Then, save the dataframe and create LaTeX:
        df_to_latex(df_desc_stat, 'df_desc_stat', caption=caption)

        - For descriptive statistics figure:
        `## Figure df_desc_stat:`
        `caption = "Descriptive statistics of ..."`
        `df_desc_stat = ...`
        `df_desc_stat['ci_low'] = df_desc_stat['mean'] - 1.96 * df_desc_stat['std'] / np.sqrt(df_desc_stat['count'])`
        `df_desc_stat['ci_high'] = ...`
        `df_to_figure(df_desc_stat, 'df_desc_stat', kind='bar', y=['mean'], y_ci=[('ci_low', 'ci_high')])`

        Note: Showing variables of different units in the same figure is not a good practice. \t
        Use a table instead.

        `# PREPROCESSING` 
        Perform any preprocessing steps needed to prepare the data for the analysis.
        For example, as applicable:
        * Creating dummy variables for categorical variables.
        * Any other data preprocessing you deem relevant.

        If no preprocessing is needed, write:
        `# No preprocessing is needed, because <your reasons here>.`

        # ANALYSIS
        Considering our "{research_goal}" and "{hypothesis_testing_plan}", decide on 1-3 additional displayitems \t
        (tables/figures) we should create for our scientific paper.
        Typically, we should have at least one displayitem for each hypothesis test.

        For each such displayitem, follow these 4 steps:
        [a] Write a comment with a unique label and add the table/figure caption. 
        Example for a table:
        `## Table df_age_death:` 
        `caption = "Test of association between age and risk of death, accounting for sex and race"`
        Example for a figure:
        `## Figure df_longevity_factors:` 
        `caption = "Adjusted and unadjusted odds ratios for longevity ..."`

        [b] Perform analysis
        - Perform appropriate analysis and/or statistical tests (see above our "{hypothesis_testing_plan}").
        - Account for relevant confounding variables, as applicable.
        - Note that you may need to perform more than one test for each hypothesis.
        - Try using inherent functionality and syntax provided in functions from the available \t
        Python packages (above).
        - Avoid, as possible, manually implementing generically available functionality. \t
        For example, to include interactions in regression analysis (if applicable), use the `formula = "y ~ a * b"` \t
        syntax in statsmodels formulas, rather than trying to manually multiply the variables.
        {mediation_note_if_applicable}\t

        [c] Create a dataframe for the scientific table/figure. 
        * Only include information that is relevant and suitable for inclusion in a scientific table/figure.
        * Nominal values should be accompanied by a measure of uncertainty (CI or STD and p-value).
        * As applicable, CI should be provided as a column of tuple (lower, upper).
        * Exclude data not important to the research goal, or that are too technical.
        * Do not repeat the same data in multiple tables/figures.
        * The df should have labels for both the columns and the index (rows): 
          - As possible, do not invent new names; just keep the original variable names from the dataset.
          - As applicable, also keep any attr names from statistical test results.

        [d] Convert the dataframe to LaTeX Table or figure using the provided functions.


        Overall, the analysis section should have the following structure:

        `# ANALYSIS`
        For each hypothesis test, create 1-3 display items (tables/figures):

        For a table:
        `## Table df_tag:`  tag is a short unique label, like 'df_age_death'
        caption = "<chosen table caption>"
        Write here code to analyze the data and create a dataframe `df_tag` for the table. 
        `df_to_latex(df_tag, 'df_tag', caption=caption)`

        For a figure:
        `## Figure df_tag:`
        caption = "<chosen figure caption>"
        Write here code to analyze the data and create a dataframe `df_tag` for the figure.
        `df_to_figure(df_tag, 'df_tag', caption=caption, kind='bar', 
        y=['height_avg', 'weight_avg', ...], 
        y_ci=['height_ci', 'weight_ci', ...],  \t
        # or y_ci=[('height_ci_low', 'height_ci_high'), ('weight_ci_low', 'weight_ci_high'), ...]
        y_p_value=['height_pval', 'weight_pval', ...])`

        etc, up to 3-4 display items in total.

        # SAVE ADDITIONAL RESULTS
        At the end of the code, after completing the tables, create a dict containing any additional \t
        results you deem important to include in the scientific paper, and save it to a pkl file \t
        'additional_results.pkl'. 

        For example:
        ```python
        additional_results = {
            'Total number of observations': <xxx>,         
            'accuracy of <mode name> model': <xxx>,
            # etc, any other results and important parameters that are not included in the displayitems
        }
        with open('additional_results.pkl', 'wb') as f:
            pickle.dump(additional_results, f)
        ```

        Avoid the following:
        Do not provide a sketch or pseudocode; write a complete runnable code including all '# HEADERS' sections.
        Do not directly use matplotlib or other plotting packages; only use the provided functions.
        Do not send any presumed output examples.
        Avoid convoluted or indirect methods of data extraction and manipulation; \t
        For clarity, use direct attribute access for clarity and simplicity.
        For clarity, access dataframes using string-based column/index names, \t
        rather than integer-based column/index positions.
        """)

    code_review_prompts: Collection[CodeReviewPrompt] = (
        CodeReviewPrompt('code flaws', None, False, dedent_triple_quote_str("""
        The code runs without any obvious bugs, but I am worried that it may have some fundamental \t
        mathematical or statistical flaws.
        I will need you to carefully check the Python code for possible flaws.
        {code_review_formatting_instructions}

        ### CHECK FOR FUNDAMENTAL FLAWS:
        Check for any fundamental mathematical or statistical flaws in the code.

        ### CHECK FOR WRONG CALCULATIONS:
        Explicitly list all key calculation in the code and look carefully for any mistakes.
        You should directly cut and paste the key calculations from the code, and carefully assess them.

        ### CHECK FOR MATH TRIVIALITIES:
        Check for any mathematically trivial assessments / statistical tests.

        ### OTHER ISSUES:
        Any other issues you find in the code.

        For each of these categories, you can provide one or more issues. 
        For example:

        ```{python_or_json}
        {
            "The analysis of <analysis name>": ["OK", "It is correct to ... "],
            "The analysis of <other analysis name>": ["CONCERN", "Forgot to include ..."],
            "The analysis of xxx vs yyy": ["CONCERN", "Different units were not standardized"],

            "mean_signal = np.mean(signal)": ("OK", "The mean is calculated correctly"],
            "sem_signal = np.std(signal)": ["CONCERN", "Forgot to divide by sqrt(n)"],
            "formula = 'y ~ a : b + c'": ["CONCERN", "The formula accounts for the interaction between a and b \t
        but does not include their main effects"],  

            "The test of positivity of mean(z)": ["CONCERN", "By definition, all z values are positive, so \t
        the mean is triviality positive"],
            "The test A > B": ["CONCERN", "In our case, this is always true because B is negative and A is positive"],
            "The test C > 0": ["OK", "This is a valid test because ..."],

            "<Any other issues you find>": ["CONCERN", "<Issue description>"],
            "<Any other assertion>": ["OK", "<Assertion description>"],
            "etc": ["OK/CONCERN", "..."],
        }
        ```

        {code_review_notes}
        """)),
        CodeReviewPrompt('data handling', None, False, dedent_triple_quote_str("""
        The code runs without any obvious bugs, but I am worried that it may contain some flaws in the analysis. 
        I will need you to carefully check the Python code for possible issues.
        {code_review_formatting_instructions}

        ### DATASET PREPARATIONS:
        - Missing values. If applicable, did we deal with missing, unknown, or undefined values,
        or with special numeric values that stand for unknown/undefined?
        Check the "{data_file_descriptions}" for any such missing values.
        - Units. If applicable, did we correctly standardize numeric values with different units 
        into same-unit values?
        - Data restriction. If applicable, are we restricting the analysis to the correct part of the data
        (based on the {hypothesis_testing_plan})?

        ### DESCRIPTIVE STATISTICS:
        As applicable, check for issues in the descriptive statistics.

        ### PREPROCESSING:
        Review the above "{data_file_descriptions}", then check our data preprocessing.
        Are we performing all needed preprocessing steps? Are we mistakenly performing any unneeded steps?

        ### ANALYSIS:        
        As applicable, check for any data analysis issues. For instance, \t
        verify that each analysis is done on the relevant data.

        ### STATISTICAL TESTS:
        Check the choice and implementation of statistical tests.

        For each of these categories, you can provide one or more issues. 
        For example:

        ```{python_or_json}
        {
            "Missing values": ["OK", "We correctly dealt with missing values"],
            "Standardizing units": ["CONCERN", "In the comparison of x and y, different units were not standardized"],
            "Data restriction": ["OK", "No data restriction is needed. We are correctly using all data"],  

            "Descriptive statistics: presented if needed": ["OK", "The code does not create a descriptive statistics \t
        table, but this is ok because ..."],
            "Descriptive statistics: variable choice": ["CONCERN", "We should not have included xxx in the table ..."],
            "Descriptive statistics: Correct data": ["CONCERN", "We mistakenly reported descriptive statistics on the \t
        data after normalization"],

            "Preprocessing": ["CONCERN", "We have normalized all variables, but xxx should not be normalized"],

            "Analysis on correct data": ["CONCERN", "We mistakenly performed the xxx analysis \t
        on the preprocessed data. This step should have been done on the original data"],

            "Choice of statistical test": ["CONCERN", "We should have used ttt test instead of sss test, because ..."],
            "Implementation of statistical test <test name>": ["OK", "The implementation is correct, because ..."],
        {regression_comments}\t
        {mediation_comments}\t
        {machine_learning_comments}\t
        }
        ```

        {code_review_notes}
        """)),
        CodeReviewPrompt('"{filename}"', 'df_*.pkl', True, dedent_triple_quote_str("""
        I ran your code.

        Here is the content of the table '{filename}' that the code created for our scientific paper:

        {file_contents_str}

        Please review the table and return a list of point-by-point assessments.

        {code_review_formatting_instructions}

        Check the following:
        ### SENSIBLE NUMERIC VALUES:
        Check each numeric value in the table and make sure it is sensible.

        ### MEASURES OF UNCERTAINTY: 
        If the table reports nominal values (like regression coefs), 
        does it also report their measures of uncertainty (like p-value, CI, or STD, as applicable)?

        ### MISSING DATA: 
        Are we missing key variables, or important results, that we should calculate and report in the table?

        ### OTHER ISSUES:
        Any other issues you find in the table.

        For each of these categories, you can provide one or more issues. 
        For example:

        ```{python_or_json}
        {
            "Sensible values": ["OK", "All the values in the table are sensible"],
            "Order of magnitude": ["CONCERN", "Weight values of 10^3 are not sensible"], 
            "CI of variables": ["CONCERN", "The CI values of 'xxx' are not flanking the mean of 'xxx'"],
            "Sign of values": ["CONCERN", "Height cannot be negative, but we have negative values"],
            "Zero values": ["CONCERN", "We have zero values for ..., but this is not possible"],

            "Measures of uncertainty": ["CONCERN", "We should have included p-values for ..."],
            "Measures of uncertainty": ["CONCERN", "CI should be provided as a column of (lower, upper) tuple"],

            "Missing data": ["CONCERN", "To fit with our hypothesis testing plan, we should have included ..."], 

            "<Any other issues you find>": ["CONCERN", "<Issue description>"],
            "<Any other assertion>": ["OK", "<Assertion description>"],
            "etc": ["OK/CONCERN", "..."],
        }
        ```

        {code_review_notes}
        """)),
        CodeReviewPrompt('all created files', '*', False, dedent_triple_quote_str("""
        I ran your code.

        Here is the content of the file(s) that the code created for our scientific paper:

        {file_contents_str}

        Please carefully review the code and these output files and return a point by point assessment.

        {code_review_formatting_instructions}:

        ### COMPLETENESS OF TABLES:
        Does the code create and output all needed results to address our {hypothesis_testing_plan}?

        ### CONSISTENCY ACROSS TABLES:
        Are the tables consistent in terms of the variables included, the measures of uncertainty, etc?

        ### MISSING DATA: 
        Are we missing key variables in a given table? Are we missing measures of uncertainty 
        (like p-value, CI, or STD, as applicable)?

        For each of these categories, you can provide one or more issues. 
        For example:

        ```{python_or_json}
        {
            "Completeness of output": ["OK", "We should include the P-values for the test in df_?.pkl"],

            "Consistency among tables": ["CONCERN", "In df_1.pkl, we provide age in years, but in df_2.pkl, \t
        we provide age in months"],

            "Missing data": ["CONCERN", "We have to add the variable 'xxx' to df_?.pkl"],
            "Measures of uncertainty": ["CONCERN", "We should have included p-values for ..."],
        {missing_tables_comments}
        }
        ```

        {code_review_notes}
        """)),
    )

    def _get_additional_contexts(self) -> Optional[Dict[str, Any]]:
        return create_pandas_and_stats_contexts(allow_dataframes_to_change_existing_series=None,
                                                enforce_saving_altered_dataframes=False,
                                                issue_if_statistics_test_not_called=True) | {
                'PickleDumpAttrReplacer': get_pickle_dump_attr_replacer(),
        }

    @staticmethod
    def _get_df_comments_for_code_and_output(code_and_output: CodeAndOutput) -> str:
        dfs = code_and_output.created_files.get_created_content_files_to_contents(
            match_filename='df_*.pkl')
        num_dfs = len(dfs)
        # is_descriptive_df = 'df_0.pkl' in dfs
        if num_dfs == 0:
            return dedent_triple_quote_str("""
                # * MISSING TABLES:
                # Note that the code does not create any tables.
                # Research papers typically have 2 or more tables. \t
                # Please suggest which tables to create and additional analysis needed.
                "Missing tables": ["CONCERN", "I suggest creating tables for ..."]""", indent=4)
        if num_dfs == 1:
            return dedent_triple_quote_str("""
                # * MISSING TABLES:
                # The code only creates 1 table.
                # Research papers typically have 2 or more tables. \t
                # Are you sure all relevant tables are created? Can you suggest any additional analysis leading \t
                to additional tables?
                "Missing tables": ["CONCERN", "I suggest creating an extra table for showing ..."]""", indent=4)
        if num_dfs == 2:
            return dedent_triple_quote_str("""
                # * MISSING TABLES:
                # Considering our research goal and hypothesis testing plan,
                # are all relevant tables created? If not, can you suggest any additional tables?
                "Missing tables": ["CONCERN", "I suggest creating an extra table showing ..."]""", indent=4)
        return ''

    @property
    def mediation_note_if_applicable(self):
        keywords = ['mediated', 'mediation', 'mediates', 'mediator', 'mediators']
        if not self.products or not self.products.hypothesis_testing_plan:
            return ""
        for hypothesis, plan in self.products.hypothesis_testing_plan['HYPOTHESES'].items():
            if any(keyword in hypothesis.lower() or keyword in plan.lower() for keyword in keywords):
                return dedent_triple_quote_str("""
                   - If you are doing a mediation analysis, don't forget to calculate both the 'a' and 'b' \t
                   paths (and add the same confounding variables to both paths, as needed).
                   """)
        return ""

    @staticmethod
    def _get_regression_comments_for_code_and_output(code_and_output: CodeAndOutput) -> str:
        if 'statsmodels' not in code_and_output.code:
            return ''
        linear_regression_funcs = ['ols(', 'OLS(', 'logit(', 'Logit(', 'glm(', 'GLM(']
        code = code_and_output.code
        func_names = [func for func in linear_regression_funcs if func in code]
        if not func_names:
            return ''
        return dedent_triple_quote_str("""\n
            # - In regressions, in case interactions terms are included:
            # Is the main effect adequately included in the model with interaction terms?
            # Did we use the `*` operator in statsmodels formula as recommended?
            # (as applicable, better use `formula = "y ~ a * b"`, instead of trying to \t
            manually multiply the variables)
            # For example:
            "Model with interaction terms": 
                ["CONCERN", "We forgot to include the main effect in the xxx model, \t
            please use the `*` operator in the formula"]
            """, indent=4)

    @staticmethod
    def _get_mediation_comments_for_code_and_output(code_and_output: CodeAndOutput) -> str:
        if 'mediation' not in code_and_output.code.lower() and False:
            return ''
        return dedent_triple_quote_str("""\n
            # - In mediation analysis:
            # did we calculate the mediation effect (e.g., using the Sobel test or other)?
            # did we account for relevant confounding factors?
            # (by adding these same confounding factors to both the 'a' and 'b' paths)
            # For example:
            "Mediation analysis":
                ["CONCERN", "We did not explicitly calculate the mediation effect"]
            """, indent=4)

    @staticmethod
    def _get_machine_learning_comments_for_code_and_output(code_and_output: CodeAndOutput) -> str:
        if 'sklearn' not in code_and_output.code:
            return ''
        ml_funcs = ['RandomForestClassifier(', 'RandomForestRegressor(',
                    'ElasticNet(', 'SVR(', 'SVC(', 'MLPRegressor(',
                    'DecisionTreeClassifier(',
                    'DecisionTreeRegressor(', 'LogisticRegression(']
        func_names = [func for func in ml_funcs if func in code_and_output.code]
        if not func_names:
            return ''
        return dedent_triple_quote_str("""\n
            # - Machine-Learning models:
            # Are we adequately performing hyperparameter tuning using cross-validation (as appropriate). 
            # Are the best hyperparameters reported (either in a table file or in the "additional_results.pkl" file).
            # For example:
            "Hyperparameter tuning":
                ["CONCERN", "We forgot to perform hyperparameter tuning"]
            """, indent=4)

    def _get_specific_attrs_for_code_and_output(self, code_and_output: CodeAndOutput) -> Dict[str, str]:
        comments = super()._get_specific_attrs_for_code_and_output(code_and_output)
        comments['regression_comments'] = self._get_regression_comments_for_code_and_output(code_and_output)
        comments['mediation_comments'] = self._get_mediation_comments_for_code_and_output(code_and_output)
        comments['machine_learning_comments'] = self._get_machine_learning_comments_for_code_and_output(code_and_output)
        comments['missing_tables_comments'] = self._get_df_comments_for_code_and_output(code_and_output)
        return comments



import matplotlib

from data_to_paper.env import DEBUG_MODE

STARTED = False


def configure_matplotlib():
    global STARTED
    if not STARTED:
        STARTED = True
        matplotlib.use('Agg')
        if DEBUG_MODE:
            print('Setting MATPLOTLIB_BACKEND')



import pandas as pd

from typing import Any, Dict, NamedTuple, Tuple, Callable, Optional


class InfoDataFrame(pd.DataFrame):
    """
    Custom DataFrame class with additional extra_info attribute
    Allows for custom state handling during pickling saving and loading
    """
    _metadata = pd.DataFrame._metadata + ['extra_info']

    def __init__(self, *args, extra_info=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.extra_info = extra_info

    # Ensure the custom constructor handles extra_info
    @property
    def _constructor(self):
        def _c(*args, **kwargs):
            cls = type(self)
            return cls(*args, extra_info=self.extra_info, **kwargs)

        return _c

    def __reduce__(self):
        pickled_state = super().__reduce__()
        # Custom state handling
        new_state = pickled_state[2].copy()
        new_state['extra_info'] = self.extra_info
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        # Restore extra_info safely
        self.extra_info = state.pop('extra_info', None)
        super().__setstate__(state)

    def transpose(self, *args, **kwargs):
        # Ensure the custom constructor handles extra_info
        df = super().transpose(*args, **kwargs)
        df = self._constructor(df)
        return df


class FuncCallParams(NamedTuple):
    """
    Stores a func call.
    """
    func: Callable
    args: Tuple[Any, ...] = ()
    kwargs: Dict[str, Any] = {}

    def call(self, **extra_kwargs):
        return self.func(*self.args, **self.kwargs, **extra_kwargs)

    @property
    def func_name(self):
        return self.func.__name__


class SaveObjFuncCallParams(FuncCallParams):
    """
    Stores a func call that saves an obj to a file
    """
    @classmethod
    def from_(cls, func, obj, filename, *args, **kwargs):
        return cls(func=func, args=(obj, filename) + args, kwargs=kwargs)

    @property
    def obj(self):
        return self.args[0]

    @property
    def filename(self):
        return self.args[1]

    @property
    def extra_args(self):
        return self.args[2:]


class InfoDataFrameWithSaveObjFuncCall(InfoDataFrame):
    """
    Custom DataFrame class where extra_info is a FuncCallParams
    """
    def __init__(self, *args, extra_info: SaveObjFuncCallParams = None, **kwargs):
        super().__init__(*args, extra_info=extra_info, **kwargs)

    def get_func_call(self) -> SaveObjFuncCallParams:
        return self.extra_info

    def get_prior_filename(self) -> Optional[str]:
        prior_df = self.extra_info.obj
        if not isinstance(prior_df, InfoDataFrameWithSaveObjFuncCall):
            return None
        return prior_df.get_func_call().filename


def save_as_func_call_df(func, df, filename, kwargs=None) -> InfoDataFrameWithSaveObjFuncCall:
    """
    save df to pickle with the func
    """
    kwargs = kwargs if kwargs is not None else {}
    df = InfoDataFrameWithSaveObjFuncCall(df, extra_info=SaveObjFuncCallParams.from_(func, df, filename, **kwargs))
    if filename:
        pickle_filename = filename + '.pkl'
        df.to_pickle(pickle_filename)
    return df



from __future__ import annotations

import inspect
from dataclasses import dataclass
from functools import wraps
from typing import Iterable, Any, get_args, get_origin, Union, Tuple

from data_to_paper.exceptions import data_to_paperException


@dataclass
class WrongTypeException(data_to_paperException, TypeError):
    """
    Raised when a value is of the wrong type.
    """
    descriptions: Tuple[str, ...] = ()
    must_be: Union[type, str, Iterable[type]] = None
    found: Union[type, str] = None

    def get_found_description(self) -> str:
        if isinstance(self.found, type):
            return f'`{name_of_type(self.found)}`'
        return self.found

    def get_must_be_description(self) -> str:
        if isinstance(self.must_be, type):
            return f'of type `{name_of_type(self.must_be)}`'
        if isinstance(self.must_be, str):
            return self.must_be
        if isinstance(self.must_be, Iterable):
            return f"of one of the types: {', '.join([name_of_type(t) for t in self.must_be])}"
        raise NotImplementedError(f"must_be: {self.must_be}")

    @property
    def found_type(self) -> type:
        return type(self.found)

    def __str__(self):
        within = ''.join([' within the ' + d for d in self.descriptions])
        return (f"object{within} must be {self.get_must_be_description()} "
                f"(but found {self.get_found_description()})")


def name_of_type(type_: type) -> str:
    """
    Get the name of the type.
    """
    return str(type_).replace('typing.', '').replace("<class '", '').replace("'>", '')


def check_all_of_type(elements: Iterable, type_: type, descriptions: Tuple[str, ...] = ()):
    """
    Check if all elements are of a certain type.
    """
    for e in elements:
        validate_value_type(e, type_, descriptions)


def check_all_of_types(elements: Iterable, types_: Iterable[type], descriptions: Tuple[str, ...] = ()):
    """
    Check if all elements are of their matching types.
    """
    for e, type_ in zip(elements, types_):
        validate_value_type(e, type_, descriptions)


def check_of_any_of_types(element: Any, types_: Iterable[type], descriptions: Tuple[str, ...] = ()):
    """
    Check if the element is of any of the types.
    """
    for type_ in types_:
        try:
            validate_value_type(element, type_, descriptions)
            return
        except WrongTypeException:
            pass
    raise WrongTypeException(descriptions=descriptions, must_be=types_, found=type(element))


def validate_value_type(value: Any, type_: type, descriptions: Tuple[str, ...] = ()):
    """
    Validate that the response is given in the correct format. if not raise WrongTypeException.
    """
    origin_type = get_origin(type_)
    if origin_type is None:
        origin_type = type_
    if origin_type is Any:
        return
    child_types = get_args(type_)

    if origin_type is Union:
        check_of_any_of_types(value, child_types, descriptions)
        return

    if origin_type.__name__ == '_empty':
        return

    if not isinstance(value, origin_type):
        raise WrongTypeException(descriptions=descriptions, must_be=origin_type, found=type(value))

    if not child_types:
        return
    if isinstance(value, dict):
        check_all_of_type(value.keys(), child_types[0], ('dict keys', ) + descriptions)
        check_all_of_type(value.values(), child_types[1], ('dict values', ) + descriptions)
    elif isinstance(value, (list, set)) and len(child_types) == 1:
        check_all_of_type(value, child_types[0], (name_of_type(type(value)), ) + descriptions)
    elif isinstance(value, tuple):
        if len(child_types) == 2 and child_types[1] is Ellipsis:
            check_all_of_type(value, child_types[0], ('tuple', ) + descriptions)
        elif len(child_types) == len(value):
            check_all_of_types(value, child_types, ('tuple', ) + descriptions)
        else:
            raise WrongTypeException(descriptions=descriptions, must_be=f'a tuple of length {len(child_types)}',
                                     found=f'a tuple of length {len(value)}')
    elif isinstance(value, Iterable):
        check_all_of_type(value, child_types[0], ('iterable', ) + descriptions)
    else:
        raise NotImplementedError(f'format_type: {type(value)} is not implemented')


def raise_on_wrong_func_argument_types(func, *args, **kwargs):
    """
    Uses the function signature to check if the arguments are of the correct type.
    Works generally on any func
    """
    sig = inspect.signature(func)
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    msgs = []
    for name, value in bound.arguments.items():
        type_ = sig.parameters[name].annotation
        try:
            validate_value_type(value, type_, (f'argument `{name}`', ))
        except WrongTypeException as e:
            msgs.append(str(e))
    if msgs:
        all_msgs = '\n'.join(msgs)
        raise TypeError(f"Error in arguments of {func.__name__}:\n{all_msgs}")


def raise_on_wrong_func_argument_types_decorator(func):
    """
    Decorator for the raise_on_wrong_func_argument_types function.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        raise_on_wrong_func_argument_types(func, *args, **kwargs)
        return func(*args, **kwargs)
    return wrapper



from data_to_paper.text.text_formatting import short_repr


def get_func_call_str(func_name: str, args, kwargs):
    return func_name + '(' + ', '.join(
        [short_repr(arg) for arg in args] + [f'{k}={short_repr(v)}' for k, v in kwargs.items()]) + ')'



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple, Dict, Type, Any, NamedTuple, Collection, List

from data_to_paper.types import HumanReviewType
from data_to_paper.env import SUPPORTED_PACKAGES, PAUSE_AT_LLM_FEEDBACK, PAUSE_AT_PROMPT_FOR_LLM_FEEDBACK, \
    AUTO_TERMINATE_AI_REVIEW, JSON_MODE, CODING_MODEL_ENGINE
from data_to_paper.interactive import PanelNames
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.run_gpt_code.run_issues import CodeProblem
from data_to_paper.code_and_output_files.output_file_requirements import TextContentOutputFileRequirement, \
    OutputFileRequirements
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.utils.replacer import Replacer, format_value
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose

from data_to_paper.interactive.human_actions import RequestInfoHumanAction
from data_to_paper.interactive.human_review import HumanReviewAppInteractor

from data_to_paper.interactive.symbols import Symbols
from data_to_paper.run_gpt_code.code_runner import CodeRunner
from data_to_paper.run_gpt_code.extract_and_check_code import CodeExtractor, ModifyAndCheckCodeExtractor
from data_to_paper.servers.model_engine import ModelEngine

from .debugger import DebuggerConverser
from .base_products_conversers import BackgroundProductsConverser
from .exceptions import FailedCreatingProductException
from .request_python_value import PythonDictReviewBackgroundProductsConverser
from .result_converser import Rewind


class CodeReviewPrompt(NamedTuple):
    name: Optional[str]

    wildcard_filename: Optional[str]
    # if None, the code review is done on the code itself without looking at the content of the created files
    # if not None, the code review is done on the content of the file(s) that match the wildcard_filename

    individually: bool
    # if True, the code review is done separately for each file that matches the wildcard_filename

    prompt: str

    # use {filename} to include the name of the created file
    # use {file_contents_str} to include the content of the created file(s)

    def get_header(self) -> str:
        header = 'Review'
        if self.name:
            header += f' of {self.name}'
        return header


@dataclass
class RequestIssuesToSolutions(PythonDictReviewBackgroundProductsConverser):
    LLM_PARAMETERS = {'temperature': 0.5}
    value_type: type = Dict[str, List[str]]
    json_mode: bool = JSON_MODE
    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a json object that can be evaluated with `json.loads()` to a Python Dict[str, List[str]].
        mapping each check you have done (keys) to a list of 2 strings, indicating either your concern \t
        ["CONCERN", "<your concern>"] or your assertion that this check is ok ["OK", "<your assertion>"].
        Like this
        ```json
        {
            "<your check>": ["CONCERN", "<your concern>"],
            "<your check>": ["OK", "<your assertion>"]
            "...": ["CONCERN/OK", "..."]
        }
        ```
        """)
    formatting_instructions_for_feedback: str = dedent_triple_quote_str("""
        Please correct your response according to my feedback, and send it again.

        Note that the different checks (dict keys) I have listed above are just examples.
        You should add/remove/modify checks to fit with the specifics of the code we are reviewing.

        Remember, your response should be formatted as {your_response_should_be_formatted_as}
        """)

    is_new_conversation: Optional[bool] = False
    rewind_after_getting_a_valid_response: Rewind = Rewind.DELETE_ALL
    default_rewind_for_result_error: Rewind = Rewind.AS_FRESH_CORRECTION

    def _check_response_value(self, response_value: Any) -> Any:
        # The type and structure of the response is checked in the parent class.
        response_value = super()._check_response_value(response_value)
        for key, (type_, feedback) in response_value.items():
            type_ = type_.upper()
            if type_ not in ('CONCERN', 'OK'):
                self._raise_self_response_error(
                    title='# Invalid value.',
                    error_message=f'The first element of the array should be "CONCERN" or "OK", but got {repr(type_)}.'
                )
            response_value[key] = (type_, feedback)
        return response_value


@dataclass
class BaseCodeProductsGPT(BackgroundProductsConverser, HumanReviewAppInteractor):
    model_engine: ModelEngine = CODING_MODEL_ENGINE
    max_code_revisions: int = 5
    max_code_writing_attempts: int = 2
    max_debug_iterations_per_attempt: int = 12
    background_product_fields_to_hide_during_code_revision: Tuple[str, ...] = ()
    debugger_cls: Type[DebuggerConverser] = DebuggerConverser
    code_and_output_cls: Type[CodeAndOutput] = CodeAndOutput
    code_extractor_cls: Type[CodeExtractor] = ModifyAndCheckCodeExtractor
    code_runner_cls: Type[CodeRunner] = CodeRunner
    supported_packages: Tuple[str, ...] = SUPPORTED_PACKAGES
    headers_required_in_code: Tuple[str, ...] = ()

    revision_round: int = 0

    provided_code: Optional[str] = None

    system_prompt: str = dedent_triple_quote_str("""
        You are a brilliant data scientist. You are writing a Python code to analyze data.
        """)

    goal_noun: str = '{code_name} code'
    goal_verb: str = 'write'
    mission_prompt: str = 'Please write a code to analyze the data.'
    termination_phrase: str = 'No issues found.'

    code_name: str = ''  # e.g. "data analysis"

    present_code_as_fresh: str = dedent_triple_quote_str("""
        Here is the code to perform the analysis.
        {created_file_names_explanation}
        ```python
        {code}
        ```
        """)  # set to None to not present code

    your_response_should_be_formatted_as: str = RequestIssuesToSolutions.your_response_should_be_formatted_as

    code_review_formatting_instructions: str = \
        'Your response should be formatted as {your_response_should_be_formatted_as}.\n' \
        'Do NOT provide any corrected code or code fragments.'

    code_review_notes: str = dedent_triple_quote_str("""
        Notes:
        (1) Generalize. 
        The points above are just examples, provided as a basis from which you can generalize to checks \t
        relevant to our specific code.
        (2) Skip irrelevant points. 
        If any of the points above are not applicable, you can skip them.
        (3) Add more checks.
        You can also add as many other relevant checks, both positive ("OK", "...") and negative ("CONCERN", "...").
        (4) Be specific.
        Be specific in the description of the checks (keys) and assessment (values), so that it is clear \t
        what you are referring to in the code.
        """)

    # set to () to skip option for revision
    code_review_prompts: Collection[CodeReviewPrompt] = (
        CodeReviewPrompt(None, '*', False, dedent_triple_quote_str("""
            I ran your code. 

            Here is the content of the output file(s) that the code created:

            {file_contents_str}

            Please check if there is anything wrong in these results (like unexpected NaN values, or anything else \t
            that may indicate that code improvements are needed).

            {code_review_formatting_instructions}

            For example:
            ```python
            {
                "NaN values in the output file":
                    ("CONCERN", "The output contains NaN values in ..."),
                "Output file should be self-contained":
                    ("CONCERN", "A header is missing for ..."),
                "Output file should contain all the required analysis": 
                    ("OK", "Nothing is missing"),
                "Sensible results": 
                    ("CONCERN", "The average of ... does not make sense"),
                "<Any other issues you find>":
                    ("CONCERN", "<Issue description>"),
                "<Any other point you checked and asserted is OK>":
                    ("OK", "<Assertion description>"),
            }
            ```

            {code_review_notes}
            """)),
    )

    _output_file_requirements: OutputFileRequirements = None

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements((
            TextContentOutputFileRequirement('results.txt'),
        ))

    @property
    def output_file_requirements(self) -> OutputFileRequirements:
        if self._output_file_requirements is None:
            self._output_file_requirements = self._create_output_file_requirements()
        return self._output_file_requirements

    @property
    def requested_output_filenames(self):
        return self.output_file_requirements.get_all_allowed_created_filenames()

    def _get_additional_contexts(self) -> Optional[Dict[str, Any]]:
        return None

    def get_created_file_names_explanation(self, code_and_output: CodeAndOutput) -> str:  # noqa
        created_files = code_and_output.created_files.get_all_created_files()
        if len(created_files) == 0:
            return ''
        elif len(created_files) == 1:
            created_file = next(iter(created_files))
            return f'It creates the file "{created_file}".'
        else:
            return f'It creates the files: {list(created_files)}.'

    def _get_specific_attrs_for_code_and_output(self, code_and_output: CodeAndOutput) -> Dict[str, str]:
        return {}

    @property
    def data_filenames(self) -> NiceList[str]:
        """
        The names of the files that gpt code can access.
        Need to be overridden by subclasses, to include the names of the data files from Products
        """
        return NiceList([],
                        wrap_with='"',
                        prefix='{} data file[s]: ')

    @property
    def data_folder(self) -> Optional[Path]:
        """
        The folder in which the data files are located.
        Need to be overridden by subclasses, to include the folder of the data files from Products
        """
        return None

    @property
    def _request_code_tag(self):
        return f'code_revision_{self.revision_round}'

    def _get_initial_code_and_output(self) -> CodeAndOutput:  # noqa
        return CodeAndOutput()

    def get_code_and_output(self) -> Optional[CodeAndOutput]:
        self.initialize_conversation_if_needed()
        code_and_output = self._get_initial_code_and_output()
        while True:
            code_and_output, debugger = self._run_debugger(code_and_output.code)
            if code_and_output is None:
                raise FailedCreatingProductException("Code debugging failed.")
            if self._get_code_review(code_and_output, debugger):
                break
            # delete created files:
            code_and_output.created_files.delete_all_created_files(self.data_folder)
            self.revision_round += 1
        code_and_output.name = self.code_name
        code_and_output.provided_code = format_value(self, self.provided_code)
        return code_and_output

    def get_debugger(self, previous_code: Optional[str] = None,
                     data_filenames: Optional[Collection[str]] = 'not_provided',
                     data_folder: Optional[Path] = 'not_provided',
                     is_new_conversation: Optional[bool] = 'not_provided',
                     ) -> DebuggerConverser:
        if data_filenames == 'not_provided':
            data_filenames = self.data_filenames
        if data_folder == 'not_provided':
            data_folder = self.data_folder
        if is_new_conversation == 'not_provided':
            is_new_conversation = False
        return self.debugger_cls.from_(
            self,
            is_new_conversation=is_new_conversation,
            max_debug_iterations=self.max_debug_iterations_per_attempt,
            background_product_fields_to_hide=(() if self.revision_round == 0
                                               else self.background_product_fields_to_hide_during_code_revision),
            code_and_output_cls=self.code_and_output_cls,
            previous_code=previous_code,
            previous_code_problem=CodeProblem.NoCode if previous_code is None else CodeProblem.AllOK,
            code_extractor_cls=self.code_extractor_cls,
            code_runner_cls=self.code_runner_cls,
            output_file_requirements=self.output_file_requirements,
            data_filenames=data_filenames,
            data_folder=data_folder,
            supported_packages=self.supported_packages,
            model_engine=self.model_engine,
            headers_required_in_code=self.headers_required_in_code,
            additional_contexts=self._get_additional_contexts(),
        )

    def _run_debugger(self, previous_code: Optional[str] = None
                      ) -> Tuple[Optional[CodeAndOutput], Optional[DebuggerConverser]]:
        for attempt in range(self.max_code_writing_attempts):
            # in each attempt, we are resetting the conversation back to this tag:
            revision_and_attempt = f"Revision {self.revision_round + 1}/{self.max_code_revisions} " \
                                   f"(attempt {attempt + 1}/{self.max_code_writing_attempts})"
            self.comment(f'Starting to write and debug code. {revision_and_attempt}.')

            # we now call the debugger that will try to run and provide feedback in multiple iterations:
            debugger = self.get_debugger(previous_code)
            code_and_output = debugger.run_debugging()
            if code_and_output is None:
                # debugging failed
                self.comment(f'Debugging failed, {revision_and_attempt}.')
                continue

            if self.present_code_as_fresh:
                # debugging succeeded. we now forge the conversation as if the LLM immediately sent the correct code:
                self._rewind_conversation_to_first_response()
                self.apply_append_surrogate_message(
                    content=Replacer(
                        self, self.present_code_as_fresh,
                        kwargs=dict(
                            code=code_and_output.code,
                            created_file_names_explanation=self.get_created_file_names_explanation(code_and_output),
                        )),
                    comment='Adding the debugged code as if it was the original response.',
                )
            return code_and_output, debugger
        return None, None

    def _get_content_files_to_contents(self, code_and_output: CodeAndOutput, wildcard_filename: str,
                                       individually: bool) -> Dict[str, str]:
        if wildcard_filename is None:
            content_files_to_contents = {None: None}
        else:
            content_files_to_contents = \
                code_and_output.created_files.get_created_content_files_to_pretty_contents(
                    view_purpose=ViewPurpose.CODE_REVIEW, match_filename=wildcard_filename, header_level=3)
            if not individually:
                content_files_to_contents = {wildcard_filename: '\n\n'.join(content_files_to_contents.values())}
        return content_files_to_contents

    @staticmethod
    def _convert_issues_to_messages(issues: Dict[str, Tuple[str, str]], header: str) -> Tuple[str, str, bool]:
        """
        Convert the issues to:
        - llm_msg: a string with the issues that are problems to send to the LLM
        - app_msg: a string with all the issues to present in the app
        - is_all_ok: True if all issues are OK, False if all issues are concerns, None if there are both
        """

        concern_issues = {issue: feedback for issue, (type_, feedback) in issues.items() if type_ != 'OK'}
        is_all_ok = True if not concern_issues else False if len(concern_issues) == len(issues) else None
        llm_msg = f'# {header}\n\n'
        if concern_issues:
            llm_msg += '\n\n'.join(f'## {issue}\n{solution}' for issue, solution in concern_issues.items())
        else:
            llm_msg += 'No issues found.'

        if issues:
            app_msg = '\n'.join(
                f'### {Symbols.get_is_ok_symbol(type_ == "OK")} {issue}\n{feedback}'
                for issue, (type_, feedback) in issues.items())
        else:
            app_msg = f'No issues found.'
        app_msg = f'## {Symbols.get_is_ok_symbol(is_all_ok)} {header}\n{app_msg}'
        return llm_msg, app_msg, is_all_ok

    def _get_llm_code_review(self, code_and_output: CodeAndOutput, auto_terminate: bool) -> str:
        if auto_terminate and self.revision_round >= self.max_code_revisions:
            return self.termination_phrase
        llm_msgs_and_is_issues: List[Tuple[str, bool]] = []
        for index, code_review_prompt in enumerate(self.code_review_prompts):
            content_files_to_contents = self._get_content_files_to_contents(
                code_and_output, code_review_prompt.wildcard_filename, code_review_prompt.individually)
            for filename, file_contents_str in content_files_to_contents.items():
                requester = RequestIssuesToSolutions.from_(
                    self,
                    model_engine=self.model_engine,
                    background_product_fields_to_hide=self.background_product_fields_to_hide_during_code_revision,
                    app=None,
                )
                replacing_kwargs = dict(
                    file_contents_str=file_contents_str,
                    filename=filename,
                    **{k: Replacer(self, v).format_text()
                       for k, v in self._get_specific_attrs_for_code_and_output(code_and_output).items()},
                )
                header = Replacer([self, requester], code_review_prompt.get_header(),
                                  kwargs=replacing_kwargs).format_text()
                formatted_code_review_prompt = \
                    Replacer([self, requester], '## Request ' + header + '\n' + code_review_prompt.prompt,
                             kwargs=replacing_kwargs).format_text()
                self._app_send_prompt(PanelNames.FEEDBACK)
                self._app_send_prompt(PanelNames.FEEDBACK, formatted_code_review_prompt,
                                      sleep_for=PAUSE_AT_PROMPT_FOR_LLM_FEEDBACK, from_md=True)
                requester.mission_prompt = formatted_code_review_prompt
                with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK,
                                                            f"Waiting for LLM {header} ({self.model_engine})"):
                    issues_to_is_ok_and_feedback = requester.run_and_get_valid_result(with_review=False)
                llm_msg, app_msg, is_all_ok = \
                    self._convert_issues_to_messages(issues_to_is_ok_and_feedback, header)
                llm_msgs_and_is_issues.append((llm_msg, is_all_ok is True))
                self._app_send_prompt(PanelNames.FEEDBACK, app_msg, sleep_for=PAUSE_AT_LLM_FEEDBACK, from_md=True)

        if all(is_ok for _, is_ok in llm_msgs_and_is_issues):
            return self.termination_phrase
        return '\n\n\n'.join(llm_msg for llm_msg, is_issues in llm_msgs_and_is_issues)

    def _get_human_code_review(self, llm_review: Optional[str] = None,
                               initial_text: str = '',
                               title: str = 'Human Code Review',
                               ) -> Optional[str]:
        if not self.app:
            return None
        human_review, human_action = self._app_receive_text_and_action(PanelNames.FEEDBACK, initial_text=initial_text,
                                                                       title=title,
                                                                       in_field_instructions=dedent_triple_quote_str("""
                Enter your feedback on code and output.
                Leave blank if no issues.
                """), optional_suggestions={'AI': llm_review, 'Default': ''})
        if isinstance(human_action, RequestInfoHumanAction):
            assert human_action.value == 'AI'
            return None
        return human_review

    def _get_code_review(self, code_and_output: CodeAndOutput,
                         debugger: DebuggerConverser) -> bool:
        """
        Return True/False indicating if the LLM wants to revise the code.
        If true, set the conversation to the state where the user ask the LLM to revise the code.
        """
        if self.actual_human_review == HumanReviewType.NONE:
            # Only LLM code review
            llm_review = self._get_llm_code_review(code_and_output, auto_terminate=True)
            human_review = None
        elif self.actual_human_review == HumanReviewType.LLM_FIRST:
            # LLM code review is performed first and sent for human review
            llm_review = self._get_llm_code_review(code_and_output, auto_terminate=AUTO_TERMINATE_AI_REVIEW)
            human_review = self._get_human_code_review(llm_review)
        elif self.actual_human_review == HumanReviewType.LLM_UPON_REQUEST:
            # LLM code review is requested only if human click "AI" button
            llm_review = None
            human_review = self._get_human_code_review(llm_review)
            if human_review is None:
                llm_review = self._get_llm_code_review(code_and_output, auto_terminate=AUTO_TERMINATE_AI_REVIEW)
                human_review = self._get_human_code_review(llm_review, title='Human Code Review (AI draft provided)',
                                                           initial_text=llm_review)
        else:
            raise ValueError(f'Invalid value for human_review: {self.actual_human_review}')

        review = llm_review if human_review is None else human_review
        is_terminating = not review.strip() or review == self.termination_phrase
        if human_review is None:
            review = review + '\n\n## Other\nPlease fix any other issues that you may find.'

        prompt_to_append_at_end_of_response = \
            Replacer(debugger, debugger.prompt_to_append_at_end_of_response).format_text()
        if not is_terminating:
            response = '# Code review\n' + dedent_triple_quote_str("""
                The code has some issues that need to be fixed:

                {code_review}

                {prompt_to_append_at_end_of_response}
                """).format(code_review=review,
                            prompt_to_append_at_end_of_response=prompt_to_append_at_end_of_response)
            self.apply_append_user_message(response)
        return is_terminating



import builtins
import openai


def serialize_exception(exception: Exception):
    return {
        'is_exception': True,
        'type': type(exception).__name__,
        'args': exception.args,
    }


def is_exception(item):
    return isinstance(item, dict) and 'is_exception' in item and item['is_exception'] is True


def de_serialize_exception(item) -> Exception:
    assert is_exception(item)
    exception_type = item['type']
    args = tuple(item['args'])
    if hasattr(builtins, exception_type):
        # if exception in builtins:
        exception = getattr(builtins, exception_type)(*args)
    elif hasattr(openai.error, exception_type):
        if exception_type == 'InvalidRequestError':
            exception = getattr(openai.error, exception_type)(*args, param=None)
        else:
            exception = getattr(openai.error, exception_type)(*args)
    else:
        exception = Exception(*args)
    return exception



from pandas import MultiIndex
from pandas.core.frame import DataFrame
from data_to_paper.latex.clean_latex import replace_special_latex_chars


def carefully_replace_special_latex_chars(s: str) -> str:
    if isinstance(s, str):
        return replace_special_latex_chars(s)
    return s


LATEX_DEFAULT_KWARGS = dict(
    escape=False,
    multicolumn=True,
    multirow=True,
    bold_rows=True,
)


def _escape_strings_in_index(index):
    if isinstance(index, MultiIndex):
        for level in range(len(index.levels)):
            index = index.set_levels(index.levels[level].map(carefully_replace_special_latex_chars), level=level)
            index.names = type(index.names)([carefully_replace_special_latex_chars(name) for name in index.names])
    index = index.map(carefully_replace_special_latex_chars)
    index.name = carefully_replace_special_latex_chars(index.name)
    return index


def _escape_string_in_dataframe(df: DataFrame) -> DataFrame:
    """
    Escape strings in a dataframe so that they can be used in latex.
    Also replace strings in the column names and index.
    use replace_special_latex_chars
    """
    for col_index in range(len(df.columns)):
        if df.iloc[:, col_index].dtype == object:
            df.iloc[:, col_index] = df.iloc[:, col_index].apply(carefully_replace_special_latex_chars)
    df.index = _escape_strings_in_index(df.index)
    df.columns = _escape_strings_in_index(df.columns)
    return df


def to_latex_with_escape(self, *args, original_method=None, on_change=None, **kwargs):
    kwargs = {**LATEX_DEFAULT_KWARGS, **kwargs}
    df = _escape_string_in_dataframe(self.copy())
    caption = kwargs.pop('caption', None)
    if caption is not None:
        caption = carefully_replace_special_latex_chars(caption)
    result = original_method(df, *args, caption=caption, **kwargs)
    return result



from copy import copy
from dataclasses import dataclass
from typing import Any

import pandas as pd

from data_to_paper.exceptions import data_to_paperException
from ..utils import format_numeric_value, temporarily_change_float_format, df_to_string_with_format_value
from ..dataframe_operations import SaveDataframeOperation, CreationDataframeOperation, \
    ChangeSeriesDataframeOperation, AddSeriesDataframeOperation, RemoveSeriesDataframeOperation


@dataclass
class BaseKeyError(KeyError, data_to_paperException):
    original_error: KeyError
    key: Any

    def __str__(self):
        return str(self.original_error)


@dataclass
class DataframeKeyError(BaseKeyError):
    available_keys: Any

    def __str__(self):
        return str(self.original_error) + \
            f"\n\nAvailable keys are:\n{self.available_keys}"


@dataclass
class DataFrameLocKeyError(BaseKeyError):
    available_column_keys: Any
    available_row_keys: Any

    def __str__(self):
        return str(self.original_error) + \
            f"\n\nAvailable row keys are:\n{self.available_row_keys}" \
            f"\n\nAvailable column keys are:\n{self.available_column_keys}"


ORIGINAL_FLOAT_FORMAT = pd.get_option('display.float_format')
TO_CSV_FLOAT_FORMAT = ORIGINAL_FLOAT_FORMAT
STR_FLOAT_FORMAT = format_numeric_value


def __init__(self, *args, created_by: str = None, file_path: str = None,
             original_method=None, on_change=None, **kwargs):
    original_method(self, *args, **kwargs)
    self.created_by = created_by
    self.file_path = file_path
    on_change(self, CreationDataframeOperation(
        id=id(self), created_by=created_by, file_path=file_path, columns=copy(self.columns.values)))


def __getitem__(self, key, original_method=None, on_change=None):
    try:
        return original_method(self, key)
    except KeyError as e:
        raise DataframeKeyError(original_error=e, key=key, available_keys=list(self.columns))


def __LocationIndexer__get_item__(self, key, *args, original_method=None, on_change=None, **kwargs):
    try:
        return original_method(self, key, *args, **kwargs)
    except KeyError as e:
        raise DataFrameLocKeyError(original_error=e, key=key,
                                   available_column_keys=list(self.obj.columns),
                                   available_row_keys=list(self.obj.index))


def __setitem__(self, key, value, original_method=None, on_change=None):
    # if value is a series, we need to check that the index is the same as the dataframe's index
    if isinstance(value, pd.Series):
        if not value.index.equals(self.index):
            raise ValueError(f"Series index ({value.index}) must be the same as dataframe index ({self.index}). "
                             f"Either drop non-matching rows, "
                             f"or use `pd.merge(..., how='outer')` to keep all rows.")

    if hasattr(self, 'columns'):
        original_columns = self.columns
    else:
        original_columns = None
    result = original_method(self, key, value)
    if original_columns is not None:
        if isinstance(key, (list, tuple)):
            is_changing_existing_columns = any(k in original_columns for k in key)
        else:
            is_changing_existing_columns = key in original_columns
        operation_type = ChangeSeriesDataframeOperation if is_changing_existing_columns else AddSeriesDataframeOperation
        on_change(self, operation_type(id=id(self), series_name=key))
    return result


def __delitem__(self, key, original_method=None, on_change=None):
    result = original_method(self, key)
    on_change(self, RemoveSeriesDataframeOperation(id=id(self), series_name=key))
    return result


def __str__(self, original_method=None, on_change=None):
    # to avoid printing with [...] skipping columns
    return self.to_string()


def to_string(self, *args, original_method=None, on_change=None, **kwargs):
    """
    We print with short floats, avoid printing with [...] skipping columns, and checking which orientation to use.
    """
    if 'float_format' in kwargs:
        numeric_formater = kwargs.pop('float_format')
    else:
        numeric_formater = STR_FLOAT_FORMAT
    return df_to_string_with_format_value(self, numeric_formater=numeric_formater, **kwargs)


def to_csv(self, *args, original_method=None, on_change=None, **kwargs):
    with temporarily_change_float_format(TO_CSV_FLOAT_FORMAT):
        result = original_method(self, *args, **kwargs)

    file_path = args[0] if len(args) > 0 else kwargs.get('path_or_buf')
    columns = copy(self.columns.values) if hasattr(self, 'columns') else None
    if file_path is not None:
        on_change(self, SaveDataframeOperation(id=id(self), file_path=file_path, columns=columns))
    return result



import re
from dataclasses import dataclass
from typing import Tuple, List, Set, Iterable

from data_to_paper.base_steps import LatexReviewBackgroundProductsConverser, \
    CheckReferencedNumericReviewBackgroundProductsConverser
from data_to_paper.base_steps.exceptions import FailedCreatingProductException
from data_to_paper.base_steps.literature_search import GET_LITERATURE_SEARCH_FOR_PRINT
from data_to_paper.latex.tables import get_displayitem_label
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.scientific_products import ScientificProducts
from data_to_paper.servers.custom_types import Citation

from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.latex.citataion_utils import find_citation_ids
from data_to_paper.utils.nice_list import nicely_join
from data_to_paper.utils.types import ListBasedSet


# TODO: need to generalize this class and move to base steps
class ShowCitationProducts:
    products: ScientificProducts = None
    background_product_fields: Tuple[str, ...] = ()

    allow_citations_from_step: str = None

    def _pre_populate_background(self):
        for content in self.get_repr_citation_products():
            self.comment(content)
        return super()._pre_populate_background()

    def get_repr_citation_products(self) -> List[str]:
        contents = []
        for product_field in self.background_product_fields:
            if product_field.startswith('literature_search') and self.products.is_product_available(product_field):
                with GET_LITERATURE_SEARCH_FOR_PRINT.temporary_set(True):
                    product = self.products[product_field]
                    contents.append(f'{product.name}:\n{product.description}')
        return contents

    def _get_available_citations(self) -> Iterable[Citation]:
        if self.allow_citations_from_step is None:
            return []
        return self.products.literature_search[self.allow_citations_from_step].get_citations()

    def _get_allowed_bibtex_citation_ids(self) -> List[str]:
        return [citation.bibtex_id for citation in self._get_available_citations()]


@dataclass
class SectionWriterReviewBackgroundProductsConverser(ShowCitationProducts,
                                                     LatexReviewBackgroundProductsConverser,
                                                     CheckReferencedNumericReviewBackgroundProductsConverser):
    """
    Base class for the writer of a paper section in latex format.
    """
    products: ScientificProducts = None
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers', 'research_goal',
                                                  'codes:data_analysis', 'latex_displayitems', 'additional_results',
                                                  'title_and_abstract')
    product_fields_from_which_response_is_extracted: Tuple[str, ...] = None
    should_remove_citations_from_section: bool = True
    un_allowed_commands: Tuple[str, ...] = (r'\verb', r'\begin{figure}')

    max_reviewing_rounds: int = 1
    goal_noun: str = '{pretty_section_names} section'
    conversation_name: str = None
    goal_verb: str = 'write'
    performer: str = 'scientific writer'
    reviewer: str = 'scientific reviewer'
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.Writer
    section_specific_instructions: str = ''
    section_review_specific_instructions: str = ''
    journal_name: str = 'Nature Communications'

    latex_instructions: str = dedent_triple_quote_str("""
        Write in tex format, escaping any math or symbols that needs tex escapes.
        """)

    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a triple-backtick "latex" block, like this:

        ```latex
        \\section{<section name>}
        <your latex-formatted writing here>
        ```
        """)

    system_prompt: str = dedent_triple_quote_str("""
        You are a data-scientist with experience writing accurate scientific research papers.

        You will write a scientific article for the journal {journal_name}, following the instructions below:
        1. Write the article section by section: Abstract, Introduction, Results, Discussion, and Methods.
        2. Write every section of the article in scientific language, in `.tex` format.
        3. Write the article in a way that is fully consistent with the scientific results we have.
        """)

    mission_prompt: str = dedent_triple_quote_str("""
        Based on the material provided above ({actual_background_product_names}), \t
        please {goal_verb} only the {goal_noun} for a {journal_name} article.
        Do not write any other parts!
        {section_specific_instructions}
        {latex_instructions}
        The {goal_noun} should be formatted as {your_response_should_be_formatted_as}
        """)

    termination_phrase: str = 'The {goal_noun} does not require any changes'

    other_system_prompt: str = dedent_triple_quote_str("""
        You are a reviewer for a scientist who is writing a scientific paper about their data analysis results.
        Your job is to provide constructive bullet-point feedback.
        We will write each section of the research paper separately. 
        If you feel that the paper section does not need further improvements, you should reply only with:
        "{termination_phrase}".
    """)

    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""\n\n
        Please correct your response according to any points in my feedback that you find relevant and applicable.
        Send back a complete rewrite of the {pretty_section_names}.
        Make sure to send the full corrected {pretty_section_names}, not just the parts that were revised.
    """)

    sentence_to_add_at_the_end_of_performer_response: str = dedent_triple_quote_str("""
        Please provide a bullet-point list of constructive feedback on the above {pretty_section_names} \t
        for my paper. Do not provide positive feedback, only provide actionable instructions for improvements in \t
        bullet points. 
        In particular, make sure that the section is correctly grounded in the information provided above.
        If you find any inconsistencies or discrepancies, please mention them explicitly in your feedback.
        {section_review_specific_instructions}

        You should only provide feedback on the {pretty_section_names}. Do not provide feedback on other sections \t
        or other parts of the paper, like LaTex Tables or Python code, provided above.

        If you don't see any flaws, respond solely with "{termination_phrase}".

        IMPORTANT: You should EITHER provide bullet-point feedback, or respond solely with "{termination_phrase}"; \t
        If you chose to provide bullet-point feedback then DO NOT include "{termination_phrase}".
        """)

    forbidden_phrases: Tuple[Tuple[str, bool], ...] = (
        # (phrase, match_case)
        ('Acknowledgments', True, ),
        ('Data Availability', False, ),
        ('Author Contributions', False, ),
        ('Competing Interests', False, ),
        ('Additional Information', False, ),
        ('References', True, ),
        ('Supplementary', False, ),
    )

    aborting_phrases: Tuple[Tuple[str, bool], ...] = (
        # (phrase, match_case)
        ('[unknown', False),
        ('<unknown', False),
        ('[insert', False),
        ('<insert', False),
        ('[missing', False),
        ('<missing', False),
        ('[to be', False),
        ('<to be', False),
        ('xx', False),
        ('xxx', False),
    )

    allow_subsections: bool = False

    def __post_init__(self):
        self.conversation_name = self.conversation_name or 'Writing: ' + \
            nicely_join([name.title() for name in self.section_names], separator=', ', last_separator=' and ')
        super().__post_init__()

    def _check_allowed_subsections(self, section: str):
        if not self.allow_subsections:
            if r'\subsection' in section:
                self._raise_self_response_error(
                    title='# Subsections are not allowed',
                    error_message='Do not include subsections in the {goal_noun}')

    @staticmethod
    def _is_pharse_in_section(section: str, phrase: str, match_case: bool) -> bool:
        if match_case:
            return phrase in section
        else:
            return phrase.lower() in section.lower()

    def _check_forbidden_phrases(self, section: str):
        used_forbidden_phrases = [
            phrase for phrase, match_case in self.forbidden_phrases
            if self._is_pharse_in_section(section, phrase, match_case)
        ]
        if used_forbidden_phrases:
            self._raise_self_response_error(
                title='# Forbidden phrases',
                error_message='Do not include: {}'.format(
                    nicely_join(used_forbidden_phrases, wrap_with='"', separator=', ')))

    def _check_for_aborting_phrases(self, section: str):
        used_aborting_phrases = [
            phrase for phrase, match_case in self.aborting_phrases
            if self._is_pharse_in_section(section, phrase, match_case)
        ]
        if used_aborting_phrases:
            raise FailedCreatingProductException(f"The LLM requires unknown values to write the section; "
                                                 f"it used the following aborting phrases:\n{used_aborting_phrases}")

    def _check_and_refine_section(self, section: str, section_name: str) -> str:
        self._check_for_aborting_phrases(section)
        section = super()._check_and_refine_section(section, section_name)
        self._check_extracted_numbers(section)
        self._check_url_in_text(section)
        self._check_forbidden_phrases(section)
        self._check_allowed_subsections(section)
        return section

    def write_sections_with_citations(self) -> List[Tuple[str, Set[Citation]]]:
        sections: List[str] = self.run_and_get_valid_result()
        sections_and_citations = []
        for section in sections:
            sections_and_citations.append(
                (section, ListBasedSet(citation for citation in self._get_available_citations()
                                       if citation.bibtex_id in find_citation_ids(section)))
            )
        return sections_and_citations


@dataclass
class FirstTitleAbstractSectionWriterReviewGPT(SectionWriterReviewBackgroundProductsConverser):
    goal_noun: str = 'title and abstract for a research paper'
    background_product_fields: Tuple[str] = ('general_dataset_description',
                                             'codes:data_analysis', 'latex_displayitems', 'additional_results')
    max_reviewing_rounds: int = 1
    conversation_name: str = 'Writing: Title and Abstract (first draft)'

    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a triple-backtick "latex" block, like this:

        ```latex
        \\title{<your latex-formatted paper title here>}

        \\begin{abstract}
        <your latex-formatted abstract here>
        \\end{abstract}
        ```
        """)

    # no need for triple quote block in title and abstract because they have clear begin-end wraps
    section_specific_instructions: str = dedent_triple_quote_str("""\n
        The Title should: 
        * be short and meaningful.
        * convey the main message, focusing on discovery not on methodology nor on the data source.
        * not include punctuation marks, such as ":,;" characters.

        The Abstract should provide a concise, interesting to read, single-paragraph summary of the paper, \t
        with the following structure:
        * short statement of the subject and its importance. 
        * description of the research gap/question/motivation.
        * short, non-technical, description of the dataset used and a non-technical explanation of the methodology.
        * summary of each of the main results. Summarize each key result which is evident from the displayitems, \t
        but without referring to specific numeric values from the displayitems.
        * statement of limitations and implications.
        """)
    section_review_specific_instructions: str = "{section_specific_instructions}"

    _raised_colon_error = False  # False to raise ":" error once. True to not raise error at all.

    def _check_and_refine_section(self, section: str, section_name: str) -> str:
        if section_name == 'title':
            if ':' in section and not self._raised_colon_error:
                self._raised_colon_error = True
                self._raise_self_response_error(
                    title='# Colon in title',
                    error_message=dedent_triple_quote_str("""
                        Titles of manuscripts in {journal_name} typically do not have a colon. '
                        Can you think of a different title that clearly state a single message without using a colon?
                        """)
                )
        if section_name == 'abstract' and section.count('\n') > 2:
            self._raise_self_response_error(
                title='# Abstract should be a single paragraph',
                error_message='The abstract should writen as a single paragraph.')
        return super()._check_and_refine_section(section, section_name)


@dataclass
class SecondTitleAbstractSectionWriterReviewGPT(FirstTitleAbstractSectionWriterReviewGPT):
    max_reviewing_rounds: int = 0
    conversation_name: str = None
    background_product_fields: Tuple[str] = ('general_dataset_description',
                                             'paper_sections:results',
                                             'literature_search:writing:background',
                                             'literature_search:writing:dataset',
                                             'literature_search:writing:results',
                                             'title_and_abstract')
    mission_prompt: str = dedent_triple_quote_str("""
        Bases on the material provided above ({actual_background_product_names}), please help me improve the \t
        title and abstract for a {journal_name} research paper. 

        {section_specific_instructions}

        I especially want you to:
        (1) Make sure that the abstract clearly states the main results of the paper \t
        (see above the {paper_sections:results}).
        (2) Make sure that the abstract correctly defines the literature gap/question/motivation \t
        (see above Literature Searches for list of related papers).

        {latex_instructions}
        Your response should be formatted as {your_response_should_be_formatted_as}
        """)


@dataclass
class IntroductionSectionWriterReviewGPT(SectionWriterReviewBackgroundProductsConverser):
    background_product_fields: Tuple[str, ...] = ('general_dataset_description', 'title_and_abstract',
                                                  'literature_search:writing:background',
                                                  'literature_search:writing:results',
                                                  'literature_search:writing:dataset',
                                                  'literature_search:writing:methods',
                                                  'paper_sections:methods',
                                                  'paper_sections:results')
    allow_citations_from_step: str = 'writing'
    should_remove_citations_from_section: bool = False
    max_reviewing_rounds: int = 1
    section_specific_instructions: str = dedent_triple_quote_str("""\n
        The introduction should be interesting and pique your reader’s interest. 
        It should be written while citing relevant papers from the Literature Searches above.

        Specifically, the introduction should follow the following multi-paragraph structure:

        * Introduce the topic of the paper and why it is important \t
        (cite relevant papers from the above "{literature_search:writing:background}"). 

        * Explain what was already done and known on the topic, and what is then the research gap/question \t
        (cite relevant papers from the above "{literature_search:writing:results}"). If there is only a minor gap, \t
        you can use language such as "Yet, it is still unclear ...", "However, less is known about ...", \t
        etc.

        * State how the current paper addresses this gap/question \t
        (cite relevant papers from the above "{literature_search:writing:dataset}" and \t
        "{literature_search:writing:results}").

        * Outline the methodological procedure and briefly state the main findings \t
        (cite relevant papers from the above "{literature_search:writing:methods}")

        Note: each of these paragraphs should be 5-6 sentence long. Do not just write short paragraphs with less \t
        than 5 sentences!  

        Citations should be added in the following format: \\cite{paper_id}.
        Do not add a \\section{References} section, I will add it later manually.

        Note that it is not advisable to write about limitations, implications, or impact in the introduction.
        """)
    section_review_specific_instructions: str = dedent_triple_quote_str("""\n
        Also, please suggest if you see any specific additional citations that are adequate to include \t
        (from the Literature Searches above).
        """)


@dataclass
class MethodsSectionWriterReviewGPT(SectionWriterReviewBackgroundProductsConverser):
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers', 'research_goal',
                                                  'codes:data_preprocessing',
                                                  'codes:data_analysis', 'title_and_abstract')
    max_reviewing_rounds: int = 0
    enforced_sub_headers: Tuple[str, ...] = ('Data Source', 'Data Preprocessing', 'Data Analysis')
    allow_subsections: bool = True

    @property
    def enforced_subheader_prompt(self) -> str:
        if self.enforced_sub_headers is None:
            return ''
        s = f'The Methods section should only have the following {len(self.enforced_sub_headers)} subsections:\n'
        for sub_header in self.enforced_sub_headers:
            s += f'* {sub_header}\n'
        return s

    section_specific_instructions: str = dedent_triple_quote_str("""\n
        The Methods section should be enclosed within triple-backtick "latex" block \
        and have 3 subsections, as follows: 

        ```latex
        \\section{Methods}

        \\subsection{Data Source}
        - Describe our data sources (see above "{data_file_descriptions}")

        \\subsection{Data Preprocessing}
        - Describe preprocessing of the data done by the Python code (see above "{codes:data_analysis}").
        - Do not include preprocessing steps that were not performed by the code. 
        - Do not include preprocessing steps that were performed by the code, but were not used as basis \t
        for further analysis affecting the result output.

        \\subsection{Data Analysis}
        - Describe each of the specific analysis steps performed by the Python code to yield the results.
        - Do not be over technical.
        - Do not enumerate the steps as a list; instead, describe the steps in a narrative form.
        ```

        Throughout the Methods section, do NOT include any of the following:
        - Missing steps not done by the code.
        - Specific version of software packages, file names, column names.
        - Names of package functions (e.g., do not say "We used sklearn.linear_model.LinearRegression", say instead \t
        "We used a linear regression model") 
        - URLs, links or references.""")

    latex_instructions: str = ''
    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a triple-backtick "latex" block, like this:
        ```latex
        \\section{Methods}
        \\subsection{Data Source}
        <your latex-formatted description of the data source here>
        \\subsection{Data Preprocessing}
        <your latex-formatted description of the data preprocessing here>
        \\subsection{Data Analysis}
        <your latex-formatted description of the data analysis here>
        ```
        """)

    section_review_specific_instructions: str = "{section_specific_instructions}"

    def _check_extracted_text_and_update_valid_result(self, extracted_text: List[str]):
        # Warn on "version = ..." :
        # e.g. "version = 1.2.3", "version 1.2.3", "Python 3.7", "Python 3.7.1"
        response = self._convert_extracted_text_to_fresh_looking_response(extracted_text)
        pattern = r'version(?:\s*=\s*|\s+)(\d+\.\d+(\.\d+)?)|Python\s+(\d+\.\d+)'
        if re.findall(pattern, response):
            self._raise_self_response_error(
                title='# Software packages',
                error_message=f'Do not mention specific version of software packages.')

        # Check subsection headings:
        pattern = r'\\subsection{([^}]*)}'
        matches = re.findall(pattern, response)
        if set(matches) != {'Data Source', 'Data Preprocessing', 'Data Analysis'}:
            self._raise_self_response_error(
                title='# Allowed subsections',
                error_message=dedent_triple_quote_str(f"""
                    The Methods section should only have the following 3 subsections: '
                    Data Source, Data Preprocessing, Data Analysis.
                    """)
            )
        return super()._check_extracted_text_and_update_valid_result(extracted_text)

    def run_and_get_valid_result(self) -> list:
        # Add code availability statement:
        response = [super().run_and_get_valid_result()[0] +
                    '\\subsection{Code Availability}\n\n'
                    'Custom code used to perform the data preprocessing and analysis, '
                    'as well as the raw code outputs, are provided in Supplementary Methods.']
        return response


@dataclass
class ResultsSectionWriterReviewGPT(SectionWriterReviewBackgroundProductsConverser):
    forbidden_phrases: Tuple[Tuple[str, bool], ...] = \
        SectionWriterReviewBackgroundProductsConverser.forbidden_phrases + \
        (
            ('In conclusions', True),
            ('Future research', False),
            ('Future work', False),
            ('Future studies', False),
            ('Future directions', False),
            ('Limitations', False),
    )
    # (phrase, match_case)

    background_product_fields: Tuple[str, ...] = \
        ('title_and_abstract', 'data_file_descriptions_no_headers_linked', 'codes:data_analysis',
         'latex_displayitems_linked', 'additional_results_linked')
    product_fields_from_which_response_is_extracted: Tuple[str, ...] = \
        ('data_file_descriptions_no_headers_linked', 'latex_displayitems_linked', 'additional_results_linked')
    self_products_to_other_products: Tuple[Tuple[str, str]] = (
        ('latex_displayitems_linked', 'latex_displayitems'),
        ('additional_results_linked', 'additional_results'),
        ('data_file_descriptions_no_headers_linked', 'data_file_descriptions_no_headers'),
    )
    max_reviewing_rounds: int = 1
    section_specific_instructions: str = dedent_triple_quote_str("""
        {general_result_instructions}
        {numeric_values_instructions}
        """)
    general_result_instructions: str = dedent_triple_quote_str("""\n
        Use the following guidelines when writing the Results:

        * Include 3-4 paragraphs, each typically focusing on one of the analysis and the resulting displayitems:
        You should typically have a separate paragraph describing each of the Tables/Figures.  \t
        If two or more display items are based on the same analysis (typically a table and a figure), \t
        they should be discussed in the same paragraph. \t
        In each such paragraph, indicate the motivation/question for the analysis, the methodology, \t
        and only then describe the results. \t
        You should describe what we see and learn from each Table/Figure. \t
        You should refer to the Tables/Figures by their labels \t
        (using \\ref{table:xxx}, \\ref{figure:xxx}), \t
        but do not add the displayitems themselves (they will be added later automatically).

        * Story-like flow: 
        It is often nice to have a story-like flow between the paragraphs, so that the reader \t
        can follow the analysis process with emphasis on the reasoning/motivation behind each analysis step. 
        For example, the first sentence of each paragraph can be a story-guiding sentences like: 
        "First, to understand whether xxx, we conducted a simple analysis of ..."; "Then, to test yyy, we performed a \t
        ..."; "Finally, to further verify the effect of zzz, we tested whether ...".

        * Conclude with a summary of the results:
        You can summarize the results at the end, with a sentence like: "In summary, these results show ...", \t
        or "Taken together, these results suggest ...".
        IMPORTANT NOTE: Your summary SHOULD NOT include a discussion of conclusions, implications, limitations, \t
        or of future work. \t
        (These will be added later as part the Discussion section, not the Results section).
        """)
    numeric_values_instructions: str = dedent_triple_quote_str("""
        * Numeric values:

        - Sources: 
        You can extract numeric values from the above provided sources: "{latex_displayitems_linked}", \t
        "{additional_results_linked}", and "{data_file_descriptions_no_headers_linked}".
        All numeric values in these sources have a \\hypertarget with a unique label. 

        - Cited numeric values should be formatted as \\hyperlink{<label>}{<value>}:
        Any numeric value extracted from the above sources should be written with a proper \\hyperlink to its \t
        corresponding source \\hypertarget.

        - Dependent values should be calculated using the \\num command.
        In scientific writing, we often need to report values which are not explicitly provided in the sources, \t
        but can rather be derived from them. For example: changing units, \t
        calculating differences, transforming regression coefficients into odds ratios, etc (see examples below).

        To derive such dependent values, please use the \\num{<formula>, "explanation"} command. 
        The <formula> contains a calculation, which will be automatically replaced with its result upon pdf compilation. 
        The "explanation" is a short textual explanation of the calculation \t
        (it will not be displayed directly in the text, but will be useful for review and traceability).  

        - Toy example for citing and calculating numeric values:

        Suppose our provided source data includes:
        ```
        No-treatment response: \\hypertarget{Z1a}{0.65} 
        With-treatment response: \\hypertarget{Z2a}{0.87}

        Treatment regression: 
        coef = \\hypertarget{Z3a}{0.17}, STD = \\hypertarget{Z3b}{0.072}, pvalue = <\\hypertarget{Z3c}{1e-6}
        ```

        Then, here are some examples of proper ways to report these provided source values:
        ```
        The no-treatment control group had a response of \\hyperlink{Z1a}{0.65} while the with-treatment \t
        group had a response of \\hyperlink{Z2a}{0.87}.

        The regression coefficient for the treatment was \\hyperlink{Z3a}{0.17} with a standard deviation of \t
        \\hyperlink{Z3b}{0.072} (P-value: < \\hyperlink{Z3c}{1e-6}).
        ```

        And here are some examples of proper ways to calculate dependent values, using the \\num command:
        ```
        The difference in response was \\num{\\hyperlink{Z2a}{0.87} - \\hyperlink{Z1a}{0.65}, \t
        "Difference between responses with and without treatment"}.

        The treatment odds ratio was \t
        \\num{exp(\\hyperlink{Z3a}{0.17}), \t
        "Translating the treatment regression coefficient to odds ratio"} (CI: \t
        \\num{exp(\\hyperlink{Z3a}{0.17} - 1.96 * \\hyperlink{Z3b}{0.072}), \t
        "low CI for treatment odds ratio, assuming normality"}, \t
        \\num{exp(\\hyperlink{Z3a}{0.17} + 1.96 * \\hyperlink{Z3b}{0.072}), \t
        "high CI for treatment odds ratio, assuming normality"}).
        ```

        * Accuracy: 
        Make sure that you are only mentioning details that are explicitly found within the Tables, Figures and \t
        Numerical Values.

        * Unknown values:
        If we need to include a numeric value that is not explicitly given in the \t
        Tables/Figures or "{additional_results_linked}", and cannot be derived from them using the \\num command, \t
        then indicate `[unknown]` instead of the numeric value. 

        For example:
        ```
        The no-treatment response was \\hyperlink{Z1a}{0.65} (STD: [unknown]).
        ```
        """)
    other_mission_prompt: str = dedent_triple_quote_str("""
        Based on the material provided above, please write the Results section for a {journal_name} research paper.

        {general_result_instructions}

        * You can use the \\num command to calculate dependent values from the provided numeric values \t
        (they will be automatically replaced with the actual numeric values in compilation).
        """)
    section_review_specific_instructions: str = dedent_triple_quote_str("""
        Do not suggest adding missing information, or stating whats missing from the displayitems or Numerical Values.
        Only suggest changes that are relevant to the Results section itself and that are supported by the given \t
        displayitems and Numerical Values.

        Do not suggest changes to the {goal_noun} that may require data not available in the the \t
        displayitems and Numerical Values.
        """)
    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""\n\n
        Please correct your response according to any points in my feedback that you find relevant and applicable.
        Send back a complete rewrite of the {pretty_section_names}.
        Make sure to send the full corrected {pretty_section_names}, not just the parts that were revised.
        Remember to include the numeric values in the format \\hyperlink{<label>}{<value>} and use the \\num command \t
        for dependent values.
    """)

    def _get_displayitem_labels(self, section_name: str) -> List[str]:
        return [get_displayitem_label(displayitem)
                for displayitem in self.products.get_latex_displayitems()[section_name]]

    def _check_and_refine_section(self, section: str, section_name: str) -> str:
        result = super()._check_and_refine_section(section, section_name)
        displayitems_labels = self._get_displayitem_labels(section_name)
        figure_or_table = 'Figure' if section_name.startswith('figure') else 'Table'
        for displayitem_label in displayitems_labels:
            if displayitem_label not in section:
                self._raise_self_response_error(
                    title='# Missing Displayitem reference',
                    error_message=dedent_triple_quote_str(f"""
                        The {section_name} section should specifically reference each of the Displayitems that we have.
                        Please make sure we have a sentence addressing {figure_or_table} "{displayitem_label}".
                        The sentence should have a reference: "{figure_or_table}~\\ref{{{displayitem_label}}}".
                        """)
                )
        return result


@dataclass
class DiscussionSectionWriterReviewGPT(SectionWriterReviewBackgroundProductsConverser):
    background_product_fields: Tuple[str, ...] = ('general_dataset_description',
                                                  'title_and_abstract',
                                                  'literature_search:writing:background',
                                                  'literature_search:writing:results',
                                                  'paper_sections:introduction',
                                                  'paper_sections:methods',
                                                  'paper_sections:results')
    allow_citations_from_step: str = 'writing'
    should_remove_citations_from_section: bool = False
    max_reviewing_rounds: int = 1
    section_review_specific_instructions: str = dedent_triple_quote_str("""\n
        Also, please suggest if you see any specific additional citations that are adequate to include \t
        (from the Literature Searches above).
        """)
    section_specific_instructions: str = dedent_triple_quote_str("""\n
        The Discussion section should follow the following structure:
        * Recap the subject of the study (cite relevant papers from the above "{literature_search:writing:background}").  
        * Recap our methodology (see "Methods" section above) and the main results \t
        (see "{paper_sections:results}" above), \t
        and compare them to the results from prior literature (see above "{literature_search:writing:results}"). 
        * Discuss the limitations of the study.
        * End with a concluding paragraph summarizing the main results, their implications and impact, \t
        and future directions.

        Citations should be added in the following format: \\cite{paper_id}.
        Do not add a \\section{References} section, I will add it later manually.
        """)



from typing import Optional

DISPLAYITEM_COMMENT_HEADER = '% This latex displayitem was generated from: '


def extract_source_filename_from_latex_displayitem(latex: str) -> Optional[str]:
    """
    Extract the source filename from a latex table/figure.
    """
    first_line = latex.split('\n')[0]
    if not first_line.startswith(DISPLAYITEM_COMMENT_HEADER):
        return None
    filename = first_line.split('`')[1]
    return filename


def embed_source_filename_as_comment_in_latex_displayitem(filename: str) -> str:
    """
    Add the source filename to a latex table/figure.
    """
    return DISPLAYITEM_COMMENT_HEADER + f'`{filename}`\n'



import numbers
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Iterable, Optional, Tuple

import numpy as np
import pandas as pd

from data_to_paper.run_gpt_code.base_run_contexts import RunContext
from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue
from data_to_paper.utils.mutable import Flag
from data_to_paper.utils.operator_value import OperatorValue

P_VALUE_MIN = 1e-6  # values smaller than this will be formatted as "<1e-6"
EPSILON = 1e-12  # 0 is formatted as "1e-12"

# We show to the LLM 0 pvalues as some small epsilon sot that it will not attempt to fix them.
# These values will be formatted as "<1e-6" in the latex tables.


class OnStr(Enum):
    RAISE = 0
    AS_FLOAT = 1
    SMALLER_THAN = 2  # with P_VALUE_MIN, e.g. "<1e-6"
    LATEX_SMALLER_THAN = 3  # with P_VALUE_MIN, e.g. "$<1e-6$"
    WITH_EPSILON = 4  # with EPSILON, e.g. "1e-12"
    WITH_ZERO = 5  # just format. no minimal value
    DEBUG = 6

    def st_sign(self):
        if self == OnStr.SMALLER_THAN:
            return '<'
        if self == OnStr.LATEX_SMALLER_THAN:
            return '$<$'
        return ''

    def minimal_value(self):
        if self in [OnStr.SMALLER_THAN, OnStr.LATEX_SMALLER_THAN]:
            return P_VALUE_MIN
        if self == OnStr.WITH_EPSILON:
            return EPSILON
        return 0


def format_p_value(x, minimal_value=P_VALUE_MIN, smaller_than_sign: str = '<'):
    """
    Format a p-value to a string.
    """
    if not isinstance(x, numbers.Number):
        return x
    if x > 1 or x < 0:
        raise ValueError(f"p-value should be in the range [0, 1]. Got: {x}")
    if np.isinf(x) or np.isnan(x):
        return str(x)
    if x >= minimal_value:
        return "{:.3g}".format(x)
    return smaller_than_sign + "{}".format(minimal_value)


@dataclass
class InvalidPValueRunIssue(RunIssue):
    """
    An issue that is raised when a p-value is invalid.
    """
    value_str: str = None
    func_call_str: str = None
    var_name: str = None
    issue: str = 'The function returned a p-value of {value_str}{for_var}.\n{call_str}'
    instructions: str = 'Please see if you understand why this is happening and fix it.'
    code_problem: CodeProblem = CodeProblem.RuntimeError

    @property
    def call_str(self):
        return f'\nThe function was called as: \n{self.func_call_str}\n\n' if self.func_call_str else ''

    @property
    def for_var(self):
        return f' for variable `{self.var_name}`' if self.var_name else ''


class PValue(OperatorValue):
    """
    An object that represents a p-value float.
    """

    # Sensible operators to perform on p-values:
    OPERATORS_RETURNING_NEW_PVALUE = {'__mul__', '__truediv__', '__floordiv__'}
    OPERATORS_RETURNING_NORMAL_VALUE = {'__eq__', '__ne__', '__lt__', '__le__', '__gt__', '__ge__',
                                        '__bool__', '__hash__'}
    # All other operators are not allowed

    BEHAVE_NORMALLY = Flag(False)
    ON_STR = OnStr.RAISE
    error_message_on_forbidden_func = "Calling `{func_name}` on a PValue object is forbidden.\n"
    this_is_a_p_value = True

    def __init__(self, value, created_by: str = None, var_name: str = None):
        super().__init__(value)
        self.created_by = created_by
        self.var_name = var_name

    def _get_new_object(self, value):
        with self.BEHAVE_NORMALLY.temporary_set(True):
            return self.__class__(value, created_by=self.created_by, var_name=self.var_name)

    def _raise_if_forbidden_func(self, method_name):
        raise RunIssue.from_current_tb(
            category='Do not convert or manipulate p-values',
            issue=self.error_message_on_forbidden_func.format(func_name=method_name, created_by=self.created_by),
            instructions='Please leave p-values intact.\n'
                         'Do not apply to p-values any function or operation '
                         '(like str, repr, format, round, operators, etc).',
            code_problem=CodeProblem.RuntimeError,
        )

    def _apply_post_operator(self, op, method_name, value):
        if self.BEHAVE_NORMALLY:
            return value
        if method_name in ['__str__', '__repr__', '__format__']:
            on_str = self.ON_STR
            if on_str == OnStr.AS_FLOAT:
                return value
            if on_str == OnStr.DEBUG:
                return f'PValue({value})'
            if on_str == OnStr.RAISE:
                self._raise_if_forbidden_func(method_name)
            return format_p_value(self.value, minimal_value=on_str.minimal_value(), smaller_than_sign=on_str.st_sign())
        if method_name in self.OPERATORS_RETURNING_NEW_PVALUE:
            return self._get_new_object(value)
        if method_name in self.OPERATORS_RETURNING_NORMAL_VALUE:
            return value
        self._raise_if_forbidden_func(method_name)

    def is_integer(self):
        return self.value.is_integer()

    @property
    def __class__(self):
        if self.BEHAVE_NORMALLY or self.ON_STR != OnStr.RAISE and self.ON_STR != OnStr.AS_FLOAT:
            return PValue
        return type(self.value)

    @classmethod
    def from_value(cls, value, created_by: str = None, var_name: str = None,
                   raise_on_nan: bool = True, raise_on_one: bool = True,
                   func_call_str: str = None, context: RunContext = None):

        if raise_on_nan and np.isnan(value):
            value_str = 'NaN'
        elif raise_on_one and value == 1:
            value_str = '1'
        else:
            value_str = None

        if value_str is not None:
            run_issue = InvalidPValueRunIssue.from_current_tb(
                category='Wrong p-value',
                value_str=value_str,
                func_call_str=func_call_str,
                var_name=var_name,
            )
            if context is None:
                raise run_issue
            context.issues.append(run_issue)

        return cls(value, created_by=created_by, var_name=var_name)

    @classmethod
    def reconstruction(cls, value, created_by: str = None, var_name: str = None):
        return cls(value, created_by=created_by, var_name=var_name)

    def __reduce__(self):
        return self.reconstruction, (self.value, self.created_by, self.var_name)


def convert_to_p_value(value, created_by: str = None, var_name: str = None,
                       raise_on_nan: bool = True, raise_on_one: bool = True,
                       func_call_str: str = None, context: RunContext = None):
    if is_p_value(value):
        return value
    kwargs = dict(created_by=created_by, var_name=var_name,
                  raise_on_nan=raise_on_nan, raise_on_one=raise_on_one,
                  func_call_str=func_call_str, context=context)
    if isinstance(value, float):
        return PValue.from_value(value, **kwargs)
    if isinstance(value, np.ndarray):
        return np.vectorize(convert_to_p_value)(value, **kwargs)
    if isinstance(value, pd.Series):
        value = value.copy()
        kwargs.pop('var_name')
        for i in range(len(value)):
            value.iloc[i] = convert_to_p_value(value.iloc[i], var_name=value.index[i], **kwargs)
        return value
    if isinstance(value, list):
        return [convert_to_p_value(val, **kwargs) for val in value]
    if isinstance(value, dict):
        return {key: convert_to_p_value(val, **kwargs) for key, val in value.items()}


def is_p_value(value):
    return hasattr(value, 'this_is_a_p_value')


def is_containing_p_value(value):
    if is_p_value(value):
        return True
    if isinstance(value, np.ndarray):
        return np.any(np.vectorize(is_containing_p_value)(value))
    if isinstance(value, pd.Series):
        return value.apply(is_containing_p_value).any()
    if isinstance(value, pd.DataFrame):
        return value.applymap(is_containing_p_value).any().any()
    if isinstance(value, (list, tuple)):
        return any(is_containing_p_value(val) for val in value)
    if isinstance(value, dict):
        return any(is_containing_p_value(val) for val in value.values())
    return False


def is_only_p_values(value):
    if is_p_value(value):
        return True
    if isinstance(value, np.ndarray):
        return np.all(np.vectorize(is_only_p_values)(value))
    if isinstance(value, pd.Series):
        return value.apply(is_only_p_values).all()
    if isinstance(value, pd.DataFrame):
        return value.applymap(is_only_p_values).all().all()
    if isinstance(value, (list, tuple)):
        return all(is_only_p_values(val) for val in value)
    if isinstance(value, dict):
        return all(is_only_p_values(val) for val in value.values())
    return False


def convert_p_values_to_floats(value):
    if is_p_value(value):
        return value.value
    if isinstance(value, np.ndarray):
        return np.vectorize(convert_p_values_to_floats)(value)
    if isinstance(value, pd.Series):
        return value.apply(convert_p_values_to_floats)
    if isinstance(value, pd.DataFrame):
        return value.applymap(convert_p_values_to_floats)
    if isinstance(value, (list, tuple)):
        return type(value)(convert_p_values_to_floats(val) for val in value)
    if isinstance(value, dict):
        return {key: convert_p_values_to_floats(val) for key, val in value.items()}
    return value


@dataclass
class TrackPValueCreationFuncs(RunContext):
    package_names: Iterable[str] = ()
    pvalue_creating_funcs: List[str] = field(default_factory=list)

    def _add_pvalue_creating_func(self, func_name: str, offset: int = 3):
        if self._is_enabled and self._is_called_from_user_script(offset=offset + 1):
            self.pvalue_creating_funcs.append(func_name)


class OnStrPValue:
    """
    A context manager for temporarily changing the ON_STR value of PValue.
    """
    def __init__(self, on_str: Optional[OnStr] = None):
        self.on_str = on_str
        self.prev_on_str = PValue.ON_STR

    def __enter__(self):
        if self.on_str is not None:
            PValue.ON_STR = self.on_str

    def __exit__(self, exc_type, exc_val, exc_tb):
        PValue.ON_STR = self.prev_on_str
        return False


@contextmanager
def pvalue_on_str_for_latex():
    """
    If the current PValue.ON_STR is SMALLER_THAN, then change it to LATEX_SMALLER_THAN.
    """
    if PValue.ON_STR == OnStr.SMALLER_THAN:
        PValue.ON_STR = OnStr.LATEX_SMALLER_THAN
        yield
        PValue.ON_STR = OnStr.SMALLER_THAN
    else:
        yield


class PValueToStars:
    default_levels = (0.01, 0.001, 0.0001)

    def __init__(self, p_value: Optional[float] = None, levels: Tuple[float] = None):
        self.p_value = p_value
        self.levels = levels or self.default_levels

    def __str__(self):
        return self.convert_to_stars()

    def convert_to_stars(self):
        p_value = self.p_value
        levels = self.levels
        if p_value < levels[2]:
            return '***'
        if p_value < levels[1]:
            return '**'
        if p_value < levels[0]:
            return '*'
        return 'ns'

    def get_conversion_legend_text(self) -> str:
        #  ns p >= 0.01, * p < 0.01, ** p < 0.001, *** p < 0.0001
        levels = self.levels
        legend = [f'ns p >= {levels[0]}']
        for i, level in enumerate(levels):
            legend.append(f'{(i + 1) * "*"} p < {level}')
        return ', '.join(legend)



from enum import Enum


class HumanReviewType(Enum):
    NONE = 'none'  # no human review
    LLM_FIRST = 'llm_first'  # LLM review is performed first and sent to human review
    LLM_UPON_REQUEST = 'llm_upon_request'  # LLM review is performed only upon human request

    def __bool__(self):
        return self is not HumanReviewType.NONE



import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import wraps
from typing import Optional, Dict, Union, Iterable, Tuple

from data_to_paper.env import REQUEST_CONTINUE_IN_PLAYBACK, FAKE_REQUEST_HUMAN_RESPONSE_ON_PLAYBACK

from data_to_paper.text.highlighted_text import format_text_with_code_blocks
from data_to_paper.utils.replacer import format_value, StrOrReplacer
from data_to_paper.text.highlighted_text import demote_html_headers
from data_to_paper.utils.mutable import Mutable

from data_to_paper.terminate.exceptions import ResetStepException

from data_to_paper.conversation.stage import Stage

from data_to_paper.servers.llm_call import get_human_response, are_more_responses_available

from .base_app import BaseApp
from .get_app import get_app
from .enum_types import PanelNames
from .human_actions import HumanAction, ButtonClickedHumanAction, TextSentHumanAction, RequestInfoHumanAction


def _skip_if_no_app(func):
    def wrapper(self, *args, **kwargs):
        if self.app is None:
            return
        return func(self, *args, **kwargs)

    return wrapper


def _raise_if_reset(func):
    """
    Add this decorator to any method where we need to check if the user has clicked a reset button.
    Only need to apply to methods that takes a while to run.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self._check_and_reset()
        result = func(self, *args, **kwargs)
        self._check_and_reset()
        return result

    return wrapper


@dataclass
class AppInteractor:

    app: Optional[BaseApp] = field(default_factory=get_app)

    def _check_and_reset(self):
        stage_to_reset_to = self._app_get_stage_to_reset_to()
        if stage_to_reset_to:
            raise ResetStepException(stage_to_reset_to)

    @_skip_if_no_app
    def _app_clear_panels(self, panel_name: Union[PanelNames, Iterable[PanelNames]] = PanelNames):
        panel_names = panel_name if isinstance(panel_name, Iterable) else [panel_name]
        for panel_name in panel_names:
            self.app.show_text(panel_name, '')
            self._app_set_panel_status(panel_name, '')

    @_skip_if_no_app
    def _app_request_panel_continue(self, panel_name: PanelNames, sleep_for: Union[None, float, bool] = 0):
        is_playback = are_more_responses_available()
        if isinstance(sleep_for, Mutable):
            sleep_for = sleep_for.val
        if sleep_for is None:
            if not is_playback or REQUEST_CONTINUE_IN_PLAYBACK:
                self.app.request_panel_continue(panel_name)
        else:
            time.sleep(sleep_for)

    @_skip_if_no_app
    def _app_send_prompt(self, panel_name: PanelNames, prompt: StrOrReplacer = '', provided_as_html: bool = False,
                         from_md: bool = False, demote_headers_by: int = 0, sleep_for: Union[None, float, bool] = 0,
                         scroll_to_bottom: bool = False):
        s = format_value(self, prompt)
        if not provided_as_html:
            do_not_format = ['latex'] if panel_name != PanelNames.PRODUCT else []
            s = format_text_with_code_blocks(s, is_html=True, width=None, from_md=from_md, do_not_format=do_not_format)
        s = demote_html_headers(s, demote_headers_by)
        self.app.show_text(panel_name, s, is_html=True, scroll_to_bottom=scroll_to_bottom)
        if isinstance(sleep_for, Mutable):
            sleep_for = sleep_for.val
        self._app_request_panel_continue(panel_name, sleep_for)

    @_skip_if_no_app
    def _app_set_focus_on_panel(self, panel_name: PanelNames):
        self.app.set_focus_on_panel(panel_name)

    def _app_receive_text_and_action(
            self, panel_name: PanelNames, initial_text: str = '',
            title: Optional[str] = '',
            instructions: Optional[str] = '',
            in_field_instructions: Optional[str] = '',
            optional_suggestions: Dict[str, str] = None,
            sleep_for: Union[None, float, bool] = 0) -> Tuple[str, HumanAction]:
        is_playback = are_more_responses_available()
        action = self._app_receive_action(panel_name, initial_text, title, instructions, in_field_instructions,
                                          optional_suggestions)
        if isinstance(action, TextSentHumanAction):
            content = action.value
        else:
            button = action.value
            if button == 'Initial':
                content = initial_text
            else:
                content = optional_suggestions[button]
        if is_playback and REQUEST_CONTINUE_IN_PLAYBACK and not FAKE_REQUEST_HUMAN_RESPONSE_ON_PLAYBACK:
            sleep_for = None  # wait for the user to click Continue
        if not is_playback:
            sleep_for = 0
        shown_text = content
        if shown_text == '' or shown_text is None:
            if isinstance(action, RequestInfoHumanAction):
                shown_text = f"Requesting {action.value}"
            else:
                shown_text = "No human feedback provided"
        self._app_send_prompt(panel_name, shown_text,
                              from_md=True, demote_headers_by=1,
                              sleep_for=sleep_for)
        return content, action

    def _app_receive_text(
            self, panel_name: PanelNames, initial_text: str = '',
            title: Optional[str] = '',
            instructions: Optional[str] = '',
            in_field_instructions: Optional[str] = '',
            optional_suggestions: Dict[str, str] = None,
            sleep_for: Union[None, float, bool] = 0) -> str:
        content, _ = self._app_receive_text_and_action(panel_name, initial_text, title, instructions,
                                                       in_field_instructions, optional_suggestions, sleep_for)
        return content

    def _app_receive_action(self, panel_name: PanelNames, initial_text: str = '',
                            title: Optional[str] = '',
                            instructions: Optional[str] = '',
                            in_field_instructions: Optional[str] = '',
                            optional_suggestions: Dict[str, str] = None) -> HumanAction:
        if self.app is None:
            return ButtonClickedHumanAction('Initial')
        return get_human_response(self.app,
                                  panel_name=panel_name,
                                  initial_text=initial_text,
                                  instructions=instructions,
                                  in_field_instructions=in_field_instructions,
                                  title=title,
                                  optional_suggestions=optional_suggestions)

    @_skip_if_no_app
    def _app_advance_stage(self, stage: Union[Stage, int, bool]):
        self.app.advance_stage(stage)

    @_skip_if_no_app
    def _app_send_product_of_stage(self, stage: Stage, product_text: str):
        self.app.send_product_of_stage(stage, product_text)

    @_skip_if_no_app
    def _app_set_panel_status(self, panel_name: PanelNames, status: str = ''):
        self.app.set_status(panel_name, 1, status)

    @contextmanager
    def _app_temporarily_set_panel_status(self, panel_name: PanelNames, status: str = ''):
        if self.app is None:
            yield
            return
        current_status = self.app.get_status(panel_name, 1)
        self._app_set_panel_status(panel_name, status)
        yield
        self._app_set_panel_status(panel_name, current_status)

    @_skip_if_no_app
    def _app_set_panel_header(self, panel_name: PanelNames, header: str):
        self.app.set_status(panel_name, 0, header)

    @_skip_if_no_app
    def _app_set_header(self, header: str, prefix: str = ''):
        self.app.set_header(prefix + header)

    @_skip_if_no_app
    def _app_send_api_usage_cost(self, stages_to_costs: Dict[Stage, float]):
        self.app.send_api_usage_cost(stages_to_costs)

    @_skip_if_no_app
    def _app_get_stage_to_reset_to(self):
        return self.app.stage_to_reset_to

    @_skip_if_no_app
    def _app_clear_stage_to_reset_to(self):
        self.app.clear_stage_to_reset_to()



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import List, Union, Callable, Tuple, Optional
from typing import TYPE_CHECKING

import openai
import tiktoken

from data_to_paper.interactive import HumanAction, BaseApp
from data_to_paper.env import CHOSEN_APP, FAKE_REQUEST_HUMAN_RESPONSE_ON_PLAYBACK, SHOW_LLM_CONTEXT
from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.utils.serialize import SerializableValue, deserialize_serializable_value
from data_to_paper.conversation.stage import Stage, delete_all_stages_following_stage

from .base_server import OrderedKeyToListServerCaller
from .model_engine import ModelEngine
from .serialize_exceptions import serialize_exception, is_exception, de_serialize_exception
from .types import InvalidAPIKeyError, ServerErrorException, MissingAPIKeyError

if TYPE_CHECKING:
    from data_to_paper.conversation.message import Message

TIME_LIMIT_FOR_OPENAI_CALL = 300  # seconds
MAX_NUM_LLM_ATTEMPTS = 5
DEFAULT_EXPECTED_TOKENS_IN_RESPONSE = 500
OPENAI_MAX_CONTENT_LENGTH_MESSAGE_CONTAINS = 'maximum context length'


# a sub-string that indicates that an openai exception was raised due to the message content being too long


@dataclass
class TooManyTokensInMessageError(Exception):
    """
    Exception raised when the number of tokens in the message is too large.
    """
    tokens: int
    expected_tokens_in_response: int
    model_engine: ModelEngine

    def __str__(self):
        return f'Number of tokens in context ({self.tokens}) is too large. ' \
               f'Expected number of tokens in response: {self.expected_tokens_in_response}. ' \
               f'Maximum number of tokens for {self.model_engine}: {self.model_engine.max_tokens}.'


@dataclass
class LLMResponse(SerializableValue):
    """
    Class to store LLM response.
    value: str - the response from the LLM.
    """


class LLMServerCaller(OrderedKeyToListServerCaller):
    """
    Class to call OpenAI API.
    """
    name = 'LLM Server'
    file_extension = '_openai.txt'
    should_log_api_cost: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_stage_callback = None
        self.api_cost_callback = None

    def set_current_stage_callback(self, callback: Optional[Callable] = None):
        self.current_stage_callback = callback

    def set_api_cost_callback(self, callback: Optional[Callable] = None):
        self.api_cost_callback = callback

    def get_current_stage(self) -> str:
        if self.current_stage_callback is not None:
            return self.current_stage_callback().value
        return "GENERAL"

    def _add_api_cost(self, cost: float):
        if self.api_cost_callback is not None:
            self.api_cost_callback(cost)

    @staticmethod
    def _get_cost_of_api_call(content: str, messages: List[Message], model_engine: ModelEngine
                              ) -> Tuple[int, int, float]:
        """
        Return the number of tokens in the input and output messages and the total cost of the API call.
        """
        tokens_in = count_number_of_tokens_in_message(messages, model_engine)
        tokens_out = count_number_of_tokens_in_message(content, model_engine)
        pricing_in, pricing_out = model_engine.pricing
        return tokens_in, tokens_out, tokens_in * pricing_in + tokens_out * pricing_out

    @staticmethod
    def _check_after_spending_money(content: str, messages: List[Message], model_engine: ModelEngine):
        tokens_in, tokens_out, cost = LLMServerCaller._get_cost_of_api_call(content, messages, model_engine)
        print_and_log_red(f'Total: {tokens_in} prompt tokens, {tokens_out} returned tokens, cost: ${cost :.2f}.',
                          should_log=False)

    def _log_api_usage_cost(self, content, messages: List[Message], model_engine: ModelEngine):
        tokens_in, tokens_out, cost = self._get_cost_of_api_call(content, messages, model_engine)
        self._add_api_cost(cost)

    def _generate_key(self, args, kwargs):
        return self.get_current_stage()

    def get_server_response(self, *args, **kwargs) -> Union[LLMResponse, HumanAction, Exception]:
        """
        returns the response from the server after post-processing. allows recording and replaying.
        """
        action = super().get_server_response(*args, **kwargs)
        if isinstance(action, str):
            action = LLMResponse(action)  # Backward compatibility
        if args[0] and self.should_log_api_cost:
            self._log_api_usage_cost(action.value, args[0], kwargs['model_engine'])
        return action

    @classmethod
    def _get_server_response(cls, messages: List[Message], model_engine: Union[ModelEngine, Callable], **kwargs
                             ) -> Union[LLMResponse, HumanAction, Exception]:
        """
        Connect with openai to get response to conversation.
        """
        if not isinstance(model_engine, ModelEngine):
            # human action:
            return model_engine(messages, **kwargs)
        print_and_log_red('Calling the LLM-API for real.', should_log=False)

        if model_engine.api_key.key is None:
            raise MissingAPIKeyError(server=cls.name, api_key=model_engine.api_key)

        openai.api_key = model_engine.api_key.key
        openai.api_base = model_engine.base_url
        for attempt in range(MAX_NUM_LLM_ATTEMPTS):
            try:
                # TODO: Need to implement timeout. Our current timeout_context() is not working on a Worker of Qt.
                response = openai.ChatCompletion.create(
                    model=model_engine.value,
                    messages=[message.to_llm_dict() for message in messages],
                    **kwargs,
                )
                break
            except openai.error.InvalidRequestError as e:
                raise ServerErrorException(server=model_engine.server_name, response=e)
            except (openai.error.AuthenticationError, openai.error.PermissionError) as e:
                raise InvalidAPIKeyError(server=model_engine.server_name, response=e, api_key=model_engine.api_key)
            except (openai.error.OpenAIError, TimeoutError) as e:
                sleep_time = 1.0 * 2 ** attempt
                print_and_log_red(f'Unexpected OPENAI error:\n{type(e)}\n{e}\n'
                                  f'Going to sleep for {sleep_time} seconds before trying again.',
                                  should_log=False)
                time.sleep(sleep_time)
                print_and_log_red(f'Retrying to call openai (attempt {attempt + 1}/{MAX_NUM_LLM_ATTEMPTS}) ...',
                                  should_log=False)
        else:
            raise ServerErrorException(
                server=model_engine.server_name,
                response=ConnectionError(f'Failed to get server response after {MAX_NUM_LLM_ATTEMPTS} attempts.'))

        content = response['choices'][0]['message']['content']
        cls._check_after_spending_money(content, messages, model_engine)
        return LLMResponse(content)

    def reset_to_stage(self, stage: Stage):
        """
        Reset the records to the records of the given stage
        """
        delete_all_stages_following_stage(self.old_records, stage)
        delete_all_stages_following_stage(self.new_records, stage)
        self.save_records()

    @staticmethod
    def _serialize_record(record: Union[SerializableValue, Exception]):
        if isinstance(record, Exception):
            return serialize_exception(record)
        if isinstance(record, SerializableValue):
            return record.serialize()
        raise ValueError(f'Cannot serialize record of type {type(record)}:\n{record}')

    @staticmethod
    def _deserialize_record(serialized_record):
        if is_exception(serialized_record):
            return de_serialize_exception(serialized_record)
        try:
            return deserialize_serializable_value(serialized_record)
        except ValueError:
            # compatible with previous versions:
            return LLMResponse(serialized_record)


OPENAI_SERVER_CALLER = LLMServerCaller()


def count_number_of_tokens_in_message(messages: Union[List[Message], str], model_engine: ModelEngine) -> int:
    """
    Count number of tokens in message using tiktoken.
    """
    if model_engine is None:
        model_engine = ModelEngine.DEFAULT
    try:
        encoding = tiktoken.encoding_for_model(model_engine.value)
    except KeyError:
        encoding = tiktoken.encoding_for_model(ModelEngine.GPT35_TURBO.value)
    if not isinstance(messages, str):
        messages = '\n'.join([message.content for message in messages])
    return len(encoding.encode(messages))


def try_get_llm_response(messages: List[Message],
                         model_engine: ModelEngine = None,
                         expected_tokens_in_response: int = None,
                         **kwargs) -> Union[str, Exception]:
    """
    Try to get a response from openai to a specified conversation.

    The conversation is sent to openai after removing comment messages and any messages indicated
    in `hidden_messages`.

    If getting a response is successful then return response string.
    If failed due to openai exception, return None.
    """
    if model_engine is None:
        model_engine = ModelEngine.DEFAULT
    if expected_tokens_in_response is None:
        expected_tokens_in_response = DEFAULT_EXPECTED_TOKENS_IN_RESPONSE
    tokens = count_number_of_tokens_in_message(messages, model_engine)
    if tokens + expected_tokens_in_response > model_engine.max_tokens:
        return TooManyTokensInMessageError(tokens, expected_tokens_in_response, model_engine)
    if SHOW_LLM_CONTEXT:
        print_and_log_red(f'Using {model_engine} (max {model_engine.max_tokens} tokens) '
                          f'for {tokens} context tokens and {expected_tokens_in_response} expected tokens.')
    else:
        print_and_log_red(f'Using {model_engine}.')
    try:
        action = OPENAI_SERVER_CALLER.get_server_response(messages, model_engine=model_engine, **kwargs)
        if isinstance(action, HumanAction):
            err = 'Human action retrieved, instead of LLM response.'
            if CHOSEN_APP == None:  # noqa (Mutable)
                err += '\nRuns recorded without human actions should be replayed with the same settings\n' \
                       '(set CHOSEN_APP to value other than None)'
            raise ValueError(err)
        assert isinstance(action, LLMResponse)
        return action.value
    except openai.error.InvalidRequestError as e:
        # TODO: add here any other exception that can be addressed by changing the number of tokens
        #     or bump up the model engine
        if OPENAI_MAX_CONTENT_LENGTH_MESSAGE_CONTAINS in str(e):
            return e
        else:
            raise


def get_human_response(app: BaseApp, **kwargs) -> HumanAction:
    """
    Allow the user to edit a message and return the edited message.
    Return None if the user did not change the message.
    """
    is_playback = are_more_responses_available()
    if FAKE_REQUEST_HUMAN_RESPONSE_ON_PLAYBACK and is_playback:
        app.request_action(**kwargs)
    response = OPENAI_SERVER_CALLER.get_server_response(
        [], model_engine=lambda messages, **k: app.request_action(**k), **kwargs)
    if isinstance(response, LLMResponse) and CHOSEN_APP != None:  # noqa (Mutable)
        raise ValueError(f'LLM response retrieved, instead of human action.\n'
                         f'Runs recorded without human actions should be replayed with the same settings\n'
                         f'(CHOSEN_APP = None)')
    assert isinstance(response, HumanAction)
    return response


def are_more_responses_available() -> bool:
    """
    Check if there are more recorded responses available from openai.
    """
    return OPENAI_SERVER_CALLER.are_more_records_available()



import contextlib
from dataclasses import dataclass
from typing import Any


@dataclass
class Mutable:
    val: Any

    def set(self, val: Any):
        self.val = val

    @contextlib.contextmanager
    def temporary_set(self, val):
        current = self.val
        self.set(val)
        try:
            yield
        finally:
            self.set(current)

    def __eq__(self, other):
        return self.val == other

    def __ne__(self, other):
        return self.val != other

    def __lt__(self, other):
        return self.val < other

    def __le__(self, other):
        return self.val <= other

    def __gt__(self, other):
        return self.val > other

    def __ge__(self, other):
        return self.val >= other

    def __bool__(self):
        return bool(self.val)


@dataclass
class Flag(Mutable):
    val: bool = False

    def __bool__(self):
        return self.val



import copy
from enum import Enum
from typing import Dict

from data_to_paper.code_and_output_files.ref_numeric_values import HypertargetPosition, HypertargetFormat


class ViewPurpose(Enum):
    FINAL_APPENDIX = 0
    FINAL_INLINE = 1
    PRODUCT = 2
    HYPERTARGET_PRODUCT = 3
    CODE_REVIEW = 4
    APP_HTML = 5

    def is_for_llm(self):
        return self in [ViewPurpose.PRODUCT, ViewPurpose.HYPERTARGET_PRODUCT, ViewPurpose.CODE_REVIEW]

    def is_for_paper(self):
        return self in [ViewPurpose.FINAL_APPENDIX, ViewPurpose.FINAL_INLINE]

    def is_for_html(self):
        return self in [ViewPurpose.APP_HTML]


DEFAULT_VIEW_PURPOSE_TO_HYPERTARGET_FORMAT: Dict[ViewPurpose, HypertargetFormat] = {
    None: HypertargetFormat(position=HypertargetPosition.NONE),
    ViewPurpose.FINAL_APPENDIX: HypertargetFormat(position=HypertargetPosition.ADJACENT, raised=True, escaped=True),
    ViewPurpose.FINAL_INLINE: HypertargetFormat(position=HypertargetPosition.ADJACENT, raised=True, escaped=False),
    ViewPurpose.PRODUCT: HypertargetFormat(position=HypertargetPosition.NONE),
    ViewPurpose.HYPERTARGET_PRODUCT: HypertargetFormat(position=HypertargetPosition.WRAP, raised=False, escaped=False),
    ViewPurpose.CODE_REVIEW: HypertargetFormat(position=HypertargetPosition.NONE),
    ViewPurpose.APP_HTML: HypertargetFormat(position=HypertargetPosition.NONE),
}


class ContentViewPurposeConverter:
    def __init__(self, view_purpose_to_params: Dict[ViewPurpose, HypertargetFormat] = None):
        view_purpose_to_params = view_purpose_to_params or DEFAULT_VIEW_PURPOSE_TO_HYPERTARGET_FORMAT
        self.view_purpose_to_params = copy.deepcopy(view_purpose_to_params)

    def convert_view_purpose_to_hypertarget_format(self, view_purpose: ViewPurpose) -> HypertargetFormat:
        return self.view_purpose_to_params[view_purpose]

    def __hash__(self):
        return hash(tuple(self.view_purpose_to_params.items()))

    def __eq__(self, other):
        return isinstance(other, ContentViewPurposeConverter) \
            and self.view_purpose_to_params == other.view_purpose_to_params



from dataclasses import dataclass
from typing import Any

from .base_products_conversers import BackgroundProductsConverser


@dataclass
class DirectorProductGPT(BackgroundProductsConverser):
    """
    Create a fake (pre-meditated) conversation, where a performer asks the Director (the application user) for a
    specific product.
    """
    system_prompt: str = None
    request_product_message: str = 'Hi, do you have a {product_name} for me?'
    provide_product_message: str = 'Yes, here is the {product_name}:\n{returned_product}\n'
    thanks_message: str = 'Thank you!'

    no_product_message: str = 'No, I do not have a {product_name} for you.'
    acknowledge_no_product_message: str = 'Ok, thank you for letting me know.'
    mission_prompt: str = None

    # inputs:
    product_name: str = None
    returned_product: Any = None

    def _get_product_from_director(self):
        """
        Ask the user for a product, such as data description, or goal.
        """
        self.apply_append_surrogate_message(self.provide_product_message, ignore=True)
        self.apply_append_user_message(self.thanks_message, ignore=True)
        return self.returned_product

    def _get_no_product(self):
        self.apply_append_surrogate_message(self.no_product_message, ignore=True)
        self.apply_append_user_message(self.acknowledge_no_product_message, ignore=True)
        return None

    def get_product_or_no_product_from_director(self, **kwargs):
        """
        Ask the user for a product, such as data description, or goal.
        """
        self.set(**kwargs)
        self.initialize_conversation_if_needed()
        self.apply_append_user_message(self.request_product_message, ignore=True)
        if self.returned_product is None:
            return self._get_no_product()
        else:
            return self._get_product_from_director()



from dataclasses import dataclass, field
from typing import Type

from data_to_paper.base_steps.base_steps_runner import DataStepRunner
from data_to_paper.conversation.stage import Stage
from data_to_paper.latex.latex_doc import LatexDocument

from .app_startup import ToyStartDialog
from .cast import DemoAgent
from .coding_steps import DemoCodeProductsGPT
from .produce_pdf_step import ProduceDemoPaperPDF
from .products import DemoProducts
from .stage import DemoStages
from .writing_steps import WriteTitleAndAbstract


@dataclass
class ToyStepsRunner(DataStepRunner):
    PROJECT_PARAMETERS_FILENAME = 'data_to_paper-toy-example.json'
    DEFAULT_PROJECT_PARAMETERS = DataStepRunner.DEFAULT_PROJECT_PARAMETERS | dict(
        research_goal=None
    )
    APP_STARTUP_CLS = ToyStartDialog

    latex_document: LatexDocument = field(default_factory=LatexDocument)

    name = 'Toy Example Research'
    stages: Type[Stage] = DemoStages
    cast = DemoAgent
    products: DemoProducts = field(default_factory=DemoProducts)

    def _pre_run_preparations(self):
        super()._pre_run_preparations()
        self.paper_producer = ProduceDemoPaperPDF.from_(
            self,
            output_filename='paper.pdf',
            paper_section_names=['title', 'abstract'],
        )

        self.stages_to_funcs = {
            DemoStages.DATA: self._run_data_step,
            DemoStages.GOAL: self._run_goal_step,
            DemoStages.CODE: self._run_code_step,
            DemoStages.WRITING: self._run_writing_step,
            DemoStages.COMPILE: self._run_compile_step,
        }

    def _run_data_step(self):
        self.products.data_file_descriptions = self.data_file_descriptions
        self.send_product_to_client('data_file_descriptions')

    def _run_goal_step(self):
        self.products.research_goal = self.project_parameters['research_goal']
        self.send_product_to_client('research_goal')

    def _run_code_step(self):
        self.products.code_and_output = DemoCodeProductsGPT.from_(self).get_code_and_output()
        self.send_product_to_client('code_and_output')

    def _run_writing_step(self):
        section_names = ['title', 'abstract']
        sections = WriteTitleAndAbstract.from_(self, section_names=section_names).run_and_get_valid_result()
        for section_name, section in zip(section_names, sections):
            self.products.paper_sections[section_name] = section
        self.send_product_to_client('title_and_abstract')

    def _run_compile_step(self):
        self.paper_producer.assemble_compile_paper()
        self._app_clear_panels()
        self._app_send_product_of_stage(
            DemoStages.COMPILE,
            f'<a href="file://{self.output_directory}/paper.pdf">Download the manuscript</a>')



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Dict, Optional

from data_to_paper.env import HUMAN_NAME

from .types import Profile, Algorithm

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from data_to_paper.conversation.conversation_actions import Action


class Agent(Enum):

    @classmethod
    @abstractmethod
    def get_primary_agent(cls) -> Agent:
        pass

    @classmethod
    @abstractmethod
    def get_human_agent(cls) -> Agent:
        pass

    @abstractmethod
    def get_conversation_name(self) -> str:
        pass

    @property
    @abstractmethod
    def profile(self) -> Profile:
        pass

    @property
    def skin_name(self) -> str:
        if self is self.get_human_agent():
            return HUMAN_NAME
        return self.profile.name

    @property
    def algorithm(self) -> Algorithm:
        return self.profile.algorithm

    @property
    def system_prompt(self) -> Optional[str]:
        return get_system_prompt(self)

    def pretty_repr(self):
        profile = self.profile
        algorithm_repr = self.algorithm.pretty_repr(self.system_prompt)
        return f"{self.skin_name} ({profile.title})\n" \
               f"{profile.description}\n" \
               f"{algorithm_repr}"

    def pretty_name(self):
        return f"{self.profile.name} ({self.value})"


AGENTS_TO_SYSTEM_PROMPTS: Dict[Agent, str] = {}


def _notify_profile_change(agent: Agent):
    """
    hook to update app client of profile change
    MAOR + ORI
    """
    pass


def set_system_prompt(agent: Agent, prompt: str):
    AGENTS_TO_SYSTEM_PROMPTS[agent] = prompt
    _notify_profile_change(agent)


def get_system_prompt(agent: Agent) -> Optional[str]:
    return AGENTS_TO_SYSTEM_PROMPTS.get(agent, None)


def on_action(action: Action):
    """
    This is called after an action was applied to a conversation.
    """
    from data_to_paper import Role
    from data_to_paper.conversation.conversation_actions import AppendMessage
    if isinstance(action, AppendMessage) and action.message.role is Role.SYSTEM:
        set_system_prompt(agent=action.message.agent, prompt=action.message.content)



from data_to_paper.interactive.base_app_startup import DataFilesStartDialog, TextEditWithHeader, MultiFileWidget


class HypothesisTestingStartDialog(DataFilesStartDialog):
    def _create_widgets(self):
        return {
            'general_description': TextEditWithHeader(
                "Dataset description", "Describe the dataset, its origin, content, purpose, etc.",
                on_change=self.update_start_button_state),
            'files_widget': MultiFileWidget(on_change=self.update_start_button_state),
            'research_goal': TextEditWithHeader(
                "Research Goal", "Specify the research goal, or leave blank for autonomous goal setting."),
        }

    def _convert_config_to_widgets(self, config):
        self.widgets['general_description'].setPlainText(config.get('general_description', ''))
        self.widgets['research_goal'].setPlainText(config.get('research_goal', '') or '')
        super()._convert_config_to_widgets(config)

    def _convert_widgets_to_config(self) -> dict:
        config = self.current_config
        config['general_description'] = self.widgets['general_description'].toPlainText() or None
        config['research_goal'] = self.widgets['research_goal'].toPlainText() or None
        config = super()._convert_widgets_to_config()
        return config

    def _get_mandatory_items_to_start(self):
        general_description_filled = bool(self.widgets['general_description'].toPlainText().strip())
        files_filled = any(file_widget.abs_path and file_widget.description_edit.toPlainText().strip()
                           for file_widget in self._get_date_files_widget().get_all_file_widgets())

        return general_description_filled, files_filled
